"""
This module supports simple linear two-port networks.

Copyright 2014--2021 Michael Hayes, UCECE
"""

from __future__ import division
from warnings import warn
import sympy as sym
from .symbols import s
from .sexpr import LaplaceDomainVoltage, LaplaceDomainCurrent, LaplaceDomainImpedance
from .sexpr import LaplaceDomainAdmittance, LaplaceDomainTransferFunction
from .sexpr import LaplaceDomainExpression
from .smatrix import LaplaceDomainVoltageMatrix, LaplaceDomainCurrentMatrix
from .smatrix import LaplaceDomainImpedanceMatrix, LaplaceDomainAdmittanceMatrix
from .cexpr import ConstantDomainExpression
from .expr import expr
from .vector import Vector
from .matrix import Matrix
from .oneport import OnePort, I, V, Y, Z
from .network import Network
from .functions import Eq, MatMul


# This needs to be generalised for superpositions.
# For now, it uses only the s-domain.

# TODO:
# 1. Defer the choice of the two-port model.  For example, a T section
# would store the three sub-networks rather than generating a B matrix.
# The appropriate model would be generated when desired.  This would
# avoid the inversion of singular matrices. The downside is that each
# object would require methods to generate each type of two-port model.
#
# Some multiport networks, such as a shunt R, have a singular Z matrix.
# Thus switching to the Y matrix and back to the Z matrix produces a
# bogus result.  The same thing occurs for a series R; this has a
# singular Y matrix.
#
# 2. Fix handling of buffered two ports (amplifier / delay).


# Consider chaining a resistor shunt to a two-port network described by the
# A matrix A1.  The shunt has an A matrix A2.  The result has an A matrix A3.
#
# A1 = [A11 A12]
#      [A21 A22]
#
# A2 = [1     0]
#      [1/R   1]
#
# A3 = A1 * A2 = [A11 + A12/R  A12]
#                [A21 + A22/R  A22]
#
# We should get the same result by adding the Y matrices.
#
# Y1 = [A22/A12  -A11 A22 / A12 + A21]
#      [-1/A12                A11/A12]
#
# Y3 = [A22/A12  -A11 A22 / A12 + A21]
#      [-1/A12          A11/A12 + 1/R]
#
# but
#
# Y2 = [inf   inf]
#      [inf   inf]
#
# We can get the correct answer using:
#
# A2 = lim x->0  [1    x]
#                [1/R  1]
#
# Now det(A2) = lim x->0 (1 - x/R)
#
# and Y2 = lim x->0 [1/x   (1/x - 1/R)]
#                   [-1/x          1/x]
#
# Note when x=0, then
#
# Y2 = lim x->0 [1/0   1/0]
#               [-1/0  1/0]
#
# and we lose the information on R.
#
# The same problem occurs with a series R and the Z matrix.
#
# A2 = lim x->0  [1    R]
#                [x    1]
#
# Now det(A2) = lim x->0 (1 - R x)
#
# and Z2 = lim x->0 [1/x  1/x - R]
#                   [1/x      1/x]
#
# Thus it is advantageous to represent two-ports by the A (or B)
# matrix.  However, things will go wrong when we transform to the Y or
# Z matrix for specific cases.

__all__ = ('Chain', 'Par2', 'Ser2', 'Hybrid2', 'InverseHybrid2',
           'Series', 'Shunt', 'IdealTransformer', 'IdealGyrator',
           'VoltageFollower', 'VoltageAmplifier',
           'IdealVoltageAmplifier', 'IdealDelay',
           'IdealVoltageDifferentiator', 'IdealVoltageIntegrator',
           'CurrentFollower', 'IdealCurrentAmplifier',
           'IdealCurrentDifferentiator', 'IdealCurrentIntegrator',
           'OpampInverter', 'OpampIntegrator', 'OpampDifferentiator',
           'TSection', 'TwinTSection', 'BridgedTSection', 'PiSection',
           'LSection', 'Ladder', 'GeneralTxLine', 'LosslessTxLine',
           'TxLine', 'AMatrix', 'BMatrix', 'GMatrix', 'HMatrix',
           'SMatrix', 'TMatrix', 'YMatrix', 'ZMatrix')

def DeltaWye(Z1, Z2, Z3):

    ZZ = (Z1 * Z2 + Z2 * Z3 + Z3 * Z1)
    return (ZZ / Z1, ZZ / Z2, ZZ / Z3)


def WyeDelta(Z1, Z2, Z3):

    ZZ = Z1 + Z2 + Z3
    return (Z2 * Z3 / ZZ, Z1 * Z3 / ZZ, Z1 * Z2 / ZZ)


def _check_oneport_args(args):

    for arg1 in args:
        if not isinstance(arg1, OnePort):
            raise ValueError('%s not a OnePort' % arg1)

class TwoPortMixin(object):

    @property
    def is_buffered(self):
        """Return true if two-port is buffered, i.e., any load
        on the output has no affect on the input. """
        # return self._A12 == 0 and self._A22 == 0
        return self._B12 == 0 and self._B22 == 0

    @property
    def is_bilateral(self):
        """Return true if two-port is bilateral. """
        return self.Bparams.det() == 1

    @property
    def is_reciprocal(self):
        """Return true if two-port is reciprocal. """
        # This also applies to Y12 == Y21, S12 == S21, or Aparams.det() == 1.
        return self._Z12 == self._Z21
    
    @property
    def is_symmetrical(self):
        """Return true if two-port is symmetrical. """
        return self._B11 == self._B22

    @property
    def is_series(self):
        """Return true if two-port is a series network. """
        # return (self._A11 == 1) and (self._A22 == 1) and (self._A21 == 0)
        return (self._B11 == 1) and (self._B22 == 1) and (self._B21 == 0)

    @property
    def is_shunt(self):
        """Return true if two-port is a shunt network. """
        # return (self._A11 == 1) and (self._A22 == 1) and (self._A12 == 0)
        return (self._B11 == 1) and (self._B22 == 1) and (self._B12 == 0)

    @property
    def _A11(self):
        """Open-circuit inverse voltage ratio"""
        return self.Aparams[0, 0]

    @property
    def _A12(self):
        """Negative short-circuit transfer impedance"""
        return self.Aparams[0, 1]

    @property
    def _A21(self):
        """Negative short-circuit inverse current ratio"""
        return self.Aparams[1, 0]

    @property
    def _A22(self):
        """Open circuit transfer admittance"""
        return self.Aparams[1, 1]

    @property
    def _B11(self):
        """Open-circuit voltage gain"""
        return self.Bparams[0, 0]

    @property
    def _B12(self):
        """Negative short-circuit transfer impedance"""
        return self.Bparams[0, 1]

    @property
    def _B21(self):
        """Negative short-circuit current gain"""
        return self.Bparams[1, 0]

    @property
    def _B22(self):
        """Open-circuit transfer admittance"""
        return self.Bparams[1, 1]

    @property
    def _G11(self):
        """Open-circuit input admittance"""
        return self.Gparams[0, 0]

    @property
    def _G12(self):
        """Short-circuit reverse current gain"""
        return self.Gparams[0, 1]

    @property
    def _G21(self):
        """Open-circuit forward voltage gain"""
        return self.Gparams[1, 0]

    @property
    def _G22(self):
        """Short-circuit output impedance"""
        return self.Gparams[1, 1]

    @property
    def _H11(self):
        """Short-circuit input impedance"""
        return self.Hparams[0, 0]

    @property
    def _H12(self):
        """Open-circuit reverse voltage gain"""
        return self.Hparams[0, 1]

    @property
    def _H21(self):
        """Short-circuit forward current gain"""
        return self.Hparams[1, 0]

    @property
    def _H22(self):
        """Open-circuit output admittance"""
        return self.Hparams[1, 1]

    @property
    def _S11(self):
        return self.Sparams[0, 0]

    @property
    def _S12(self):
        return self.Sparams[0, 1]

    @property
    def _S21(self):
        return self.Sparams[1, 0]

    @property
    def _S22(self):
        return self.Sparams[1, 1]

    @property
    def _T11(self):
        return self.Tparams[0, 0]

    @property
    def _T12(self):
        return self.Tparams[0, 1]

    @property
    def _T21(self):
        return self.Tparams[1, 0]

    @property
    def _T22(self):
        return self.Tparams[1, 1]        

    @property
    def _Y11(self):
        """Short-circuit input admittance"""
        return self.Yparams[0, 0]

    @property
    def _Y12(self):
        """Short-circuit reverse transfer admittance"""
        return self.Yparams[0, 1]

    @property
    def _Y21(self):
        """Short-circuit transfer admittance"""
        return self.Yparams[1, 0]

    @property
    def _Y22(self):
        """Short-circuit output admittance"""
        return self.Yparams[1, 1]

    @property
    def _Z11(self):
        """Open-cicuit input impedance"""
        return self.Zparams[0, 0]

    @property
    def _Z12(self):
        """Open-cicuit transfer impedance"""
        return self.Zparams[0, 1]

    @property
    def _Z21(self):
        """Open-cicuit reverse transfer impedance"""
        return self.Zparams[1, 0]

    @property
    def _Z22(self):
        """Open-cicuit output impedance"""
        return self.Zparams[1, 1]

    @property
    def A11(self):
        """Open-circuit inverse voltage ratio"""
        return expr(self._A11)

    @property
    def A12(self):
        """Negative short-circuit transfer impedance"""
        return expr(self._A12)

    @property
    def A21(self):
        """Negative short-circuit inverse current ratio"""
        return expr(self._A21)

    @property
    def A22(self):
        """Open circuit transfer admittance"""
        return expr(self._A22)

    @property
    def B11(self):
        """Open-circuit voltage gain"""
        return expr(self._B11)

    @property
    def B12(self):
        """Negative short-circuit transfer impedance"""
        return expr(self._B12)

    @property
    def B21(self):
        """Negative short-circuit current gain"""
        return expr(self._B21)

    @property
    def B22(self):
        """Open-circuit transfer admittance"""
        return expr(self._B22)

    @property
    def G11(self):
        """Open-circuit input admittance"""
        return expr(self._G11)

    @property
    def G12(self):
        """Short-circuit reverse current gain"""
        return expr(self._G12)

    @property
    def G21(self):
        """Open-circuit forward voltage gain"""
        return expr(self._G21)

    @property
    def G22(self):
        """Short-circuit output impedance"""
        return expr(self._G22)

    @property
    def H11(self):
        """Short-circuit input impedance"""
        return expr(self._H11)

    @property
    def H12(self):
        """Open-circuit reverse voltage gain"""
        return expr(self._H12)

    @property
    def H21(self):
        """Short-circuit forward current gain"""
        return expr(self._H21)

    @property
    def H22(self):
        """Open-circuit output admittance"""
        return expr(self._H22)

    @property
    def S11(self):
        """S11"""
        return expr(self._S11)

    @property
    def S12(self):
        """S12"""        
        return expr(self._S12)

    @property
    def S21(self):
        """S21"""        
        return expr(self._S21)

    @property
    def S22(self):
        """S22"""        
        return expr(self._S22)

    @property
    def T11(self):
        """T11"""        
        return expr(self._T11)

    @property
    def T12(self):
        """T12"""        
        return expr(self._T12)

    @property
    def T21(self):
        """T21"""        
        return expr(self._T21)

    @property
    def T22(self):
        """T22"""
        return expr(self._T22)

    @property
    def Y11(self):
        """Short-circuit input admittance"""
        return expr(self._Y11)

    @property
    def Y12(self):
        """Short-circuit reverse transfer admittance"""
        return expr(self._Y12)

    @property
    def Y21(self):
        """Short-circuit transfer admittance"""
        return expr(self._Y21)

    @property
    def Y22(self):
        """Short-circuit output admittance"""
        return expr(self._Y22)

    @property
    def Z11(self):
        """Open-cicuit input impedance"""
        return expr(self._Z11)

    @property
    def Z12(self):
        """Open-cicuit transfer impedance"""
        return expr(self._Z12)

    @property
    def Z21(self):
        """Open-cicuit reverse transfer impedance"""
        return expr(self._Z21)

    @property
    def Z22(self):
        """Open-cicuit output impedance"""
        return expr(self._Z22)
        

class TwoPortMatrix(Matrix, TwoPortMixin):

    # The following default properties are fallbacks when other conversions have
    # not been defined.

    @property
    def Aparams(self):
        return AMatrix(self.Bparams.inv())

    @property
    def Bparams(self):
        return BMatrix(self.Aparams.inv())

    @property
    def Gparams(self):
        return GMatrix(self.Hparams.inv())

    @property
    def Hparams(self):
        return HMatrix(self.Gparams.inv())

    @property
    def Sparams(self):
        return self.Aparams.Sparams

    @property
    def Tparams(self):
        return self.Sparams.Tparams

    @property
    def Yparams(self):
        return YMatrix(self.Zparams.inv())

    @property
    def Zparams(self):
        return ZMatrix(self.Yparams.inv())

    def pdb(self):
        import pdb; pdb.set_trace()
        return self
    

class AMatrix(TwoPortMatrix):

    """A-parameters (ABCD parameters, chain matrix)
    ::
       +-  -+     +-       -+   +-  -+
       | V1 |  =  | A11  A12|   | V2 |
       | I1 |     | A21  A22|   |-I2 |
       +-  -+     +-       -+   +-  -+

              +-         -+
       units  | 1     ohm |
              | 1/ohm   1 |
              +-         -+

    A buffered two-port has A12 = A22 = 0.

    A = inv(B)
    """

    @classmethod
    def generic(cls):
        return cls((('A_11', 'A_12'), ('A_21', 'A_22')))        

    def equation(self):
        return Eq(Matrix(('V1', 'I1')), MatMul(self, Matrix(('V2', '-I2'))),
                  evaluate=False)
    
    @property
    def Aparams(self):
        # Perhaps we should make a copy?
        return self

    @property
    def Bparams(self):

        # Inverse
        det = self.det().expr
        if det == 0:
            warn('Producing dodgy B matrix')
        return BMatrix(((self._A22 / det, -self._A12 / det),
                        (-self._A21 / det, self._A11 / det)))

    @property
    def Hparams(self):

        if self._A22 == 0:
            warn('Producing dodgy H matrix')
        det = self.det().expr            
        return HMatrix(((self._A12 / self._A22, det / self._A22),
                        (-1 / self._A22, self._A21 / self._A22)))


    @property
    def Sparams(self):
        Z0 = LaplaceDomainImpedance('Z_0').as_expr()
        d = self._A12 + Z0 * (self._A11 + self._A22) + Z0**2 * self._A21
        return SMatrix((((self._A12 + Z0 * (self._A11 - self._A22) - Z0**2 * self._A21) / d,
                         (2 * Z0 * (self._A11 * self._A22 - self._A12 * self._A21)) / d),
                         (2 * Z0 / d, (self._A12 - Z0 * (self._A11 - self._A22) - Z0**2 * self._A21) / d)))


    @property
    def Yparams(self):

        # This produces a bogus Y matrix when A12 is zero (say for a
        # shunt element).   Note, it doesn't use A21.
        if self._A12 == 0:
            warn('Producing dodgy Y matrix')
        det = self.det().expr            
        return YMatrix(((self._A22 / self._A12, -det / self._A12),
                        (-1 / self._A12, self._A11 / self._A12)))

    @property
    def Zparams(self):

        # This produces a bogus Z matrix when A21 is zero (say for a
        # series element).   Note, it doesn't use A12.
        if self._A21 == 0:
            warn('Producing dodgy Z matrix')
        det = self.det().expr            
        return ZMatrix(((self._A11 / self._A21, det / self._A21),
                        (1 / self._A21, self._A22 / self._A21)))

    @property
    def Z1oc(self):
        """open-circuit input impedance"""
        # Z11
        return LaplaceDomainImpedance(self._A11 / self._A21)

    @classmethod
    def Zseries(cls, Zval):

        if not isinstance(Zval, LaplaceDomainImpedance):
            raise ValueError('Zval not LaplaceDomainImpedance')

        return cls(((1, Zval),
                    (0, 1)))

    @classmethod
    def Yseries(cls, Yval):

        if not isinstance(Yval, LaplaceDomainAdmittance):
            raise ValueError('Yval not LaplaceDomainAdmittance')

        return cls(((1, 1 / Yval),
                    (0, 1)))

    @classmethod
    def Yshunt(cls, Yval):

        if not isinstance(Yval, LaplaceDomainAdmittance):
            raise ValueError('Yval not LaplaceDomainAdmittance')

        return cls(((1, 0),
                    (Yval, 1)))

    @classmethod
    def Zshunt(cls, Zval):

        if not isinstance(Zval, LaplaceDomainImpedance):
            raise ValueError('Zval not LaplaceDomainImpedance')

        return cls(((1, 0),
                    (1 / Zval, 1)))

    @classmethod
    def transformer(cls, alpha):

        alpha = ConstantDomainExpression(alpha)

        return cls(((1 / alpha, 0),
                    (0, alpha)))

    @classmethod
    def gyrator(cls, R):

        R = ConstantDomainExpression(R)

        return cls(((0, R),
                    (1 / R, 0)))

    @classmethod
    def Lsection(cls, Z1, Z2):

        return cls.Zseries(Z1).chain(cls.Zshunt(Z2))

    @classmethod
    def Tsection(cls, Z1, Z2, Z3):

        return cls.Lsection(Z1, Z2).chain(cls.Zseries(Z3))

    @classmethod
    def Pisection(cls, Z1, Z2, Z3):

        return cls.Zshunt(Z1).chain(cls.Lsection(Z2, Z3))

    def chain(self, OP):

        return self * OP

    def cascade(self, OP):

        return self.chain(OP)


class BMatrix(TwoPortMatrix):

    """B-parameters (inverse ABCD parameters)
    ::
       +-  -+     +-       -+   +-  -+
       | V2 |  =  | B11  B12|   | V1 |
       |-I2 |     | B21  B22|   | I1 |
       +-  -+     +-       -+   +-  -+

              +-         -+
       units  | 1     ohm |
              | 1/ohm   1 |
              +-         -+

    B = inv(A)
    """

    @classmethod
    def generic(cls):
        return cls((('B_11', 'B_12'), ('B_21', 'B_22')))

    def equation(self):
        return Eq(Matrix(('V2', '-I2')), MatMul(self, Matrix(('V1', 'I1'))),
                  evaluate=False)
        
    @property
    def Aparams(self):
        # Inverse
        det = self.det().expr
        return AMatrix(((self._B22 / det, -self._B12 / det),
                        (-self._B21 / det, self._B11 / det)))

    @property
    def Bparams(self):
        # Perhaps we should make a copy?
        return self

    @property
    def Gparams(self):

        det = self.det().expr        
        return GMatrix(((-self._B21 / self._B22, -1 / self._B22),
                        (det / self._B22, -self._B12 / self._B22)))

    @property
    def Hparams(self):

        det = self.det().expr        
        return HMatrix(((-self._B12 / self._B11, 1 / self._B11),
                        (-det / self._B11, -self._B21 / self._B11)))

    @property
    def Yparams(self):

        det = self.det().expr
        return YMatrix(((-self._B11 / self._B12, 1 / self._B12),
                        (det / self._B12, -self._B22 / self._B12)))

    @property
    def Zparams(self):

        det = self.det().expr        
        return ZMatrix(((-self._B22 / self._B21, -1 / self._B21),
                        (-det / self._B21, -self._B11 / self._B21)))

    @property
    def Z1oc(self):
        """open-circuit input impedance"""
        # Z11
        return LaplaceDomainImpedance(-self._B22 / self._B21)

    @classmethod
    def Zseries(cls, Zval):

        if not isinstance(Zval, LaplaceDomainImpedance):
            raise ValueError('Zval not LaplaceDomainImpedance')

        return cls(((1, -Zval),
                    (0, 1)))

    @classmethod
    def Yseries(cls, Yval):

        if not isinstance(Yval, LaplaceDomainAdmittance):
            raise ValueError('Yval not LaplaceDomainAdmittance')

        return cls(((1, -1 / Yval),
                    (0, 1)))

    @classmethod
    def Yshunt(cls, Yval):

        if not isinstance(Yval, LaplaceDomainAdmittance):
            raise ValueError('Yval not LaplaceDomainAdmittance')

        return cls(((1, 0),
                    (-Yval, 1)))

    @classmethod
    def Zshunt(cls, Zval):

        if not isinstance(Zval, LaplaceDomainImpedance):
            raise ValueError('Zval not LaplaceDomainImpedance')

        return cls(((1, 0),
                    (-1 / Zval, 1)))

    @classmethod
    def voltage_amplifier(cls, Af, Ar=1e-9, Yin=1e-9, Zout=1e-9):
        """Voltage amplifier
        Af forward voltage gain
        Ar reverse voltage gain (ideally 0)
        Yin input admittance (ideally 0)
        Zout output impedance (ideally 0)
        """

        if Ar == 0 and Yin == 0 and Zout == 0:
            warn('Should use G matrix; tweaking B matrix to make invertible')
            Ar = 1e-9
            Yin = 1e-9
            Zout = 1e-9

        Af = LaplaceDomainExpression(Af)
        Ar = LaplaceDomainExpression(Ar)
        Yin = LaplaceDomainExpression(Yin)
        Zout = LaplaceDomainExpression(Zout)

        # This should be defined with a G matrix
        #
        # G = [0   0]
        #     [Af  0]
        #
        # With this model, the inverse voltage gain is 1 / Af
        #
        # G = lim x->0  [0   x]
        #               [Af  0]
        #
        # B = lim x->0  [Af    0/x]
        #               [0/x  -1/x]
        #
        # A = lim x->0  [1/Af 0/Af]
        #               [0/Af   -x]

        # Perhaps default Ar, Yin, and Zout to 1e-10 to get a reasonable
        # B matrix?

        return cls(((1 / Ar, -1 / (Ar * Yin),
                    ( -1 / (Ar * Zout)), -1 / (Ar * Yin * Zout * (Af * Ar - 1)))))

    @classmethod
    def current_amplifier(cls, Af, Ar=1e-9, Zin=1e-9, Yout=1e-9):
        """Current amplifier
        Af forward current gain
        Ar reverse current gain (ideally 0)
        Yin input admittance (ideally 0)
        Yout output impedance (ideally 0)
        """

        if Ar == 0 and Zin == 0 and Yout == 0:
            warn('Should use G matrix; tweaking B matrix to make invertible')
            Ar = 1e-9
            Zin = 1e-9
            Yout = 1e-9

        Af = LaplaceDomainExpression(Af)
        Ar = LaplaceDomainExpression(Ar)
        Zin = LaplaceDomainExpression(Zin)
        Yout = LaplaceDomainExpression(Yout)

        # This should be defined with a H matrix
        #
        # H = [0   0]
        #     [Af  0]
        #
        # With this model, the inverse current gain is 1 / Af
        #
        # H = lim x->0  [0   x]
        #               [Af  0]
        #
        # B = lim x->0  [1/x    0]
        #               [0    -Af]
        #
        # A = lim x->0  [1/x  -0/x]
        #               [0/x   -Af]

        return cls(((1 / Ar, -1 / (Ar * Yout)),
                    (-1 / (Ar * Zin), -1 / (Ar * Yout * Zin * (Af * Ar - 1)))))

    @classmethod
    def voltage_differentiator(cls, Av=1):

        return cls.voltage_amplifier(LaplaceDomainExpression(Av).differentiate())

    @classmethod
    def voltage_integrator(cls, Av):

        return cls.voltage_amplifier(LaplaceDomainExpression(Av).integrate())

    @classmethod
    def current_differentiator(cls, Av):

        return cls.current_amplifier(LaplaceDomainExpression(Av).differentiate())

    @classmethod
    def current_integrator(cls, Av):

        return cls.current_amplifier(LaplaceDomainExpression(Av).integrate())

    @classmethod
    def transformer(cls, alpha):

        alpha = ConstantDomainExpression(alpha)

        return cls(((alpha, 0),
                    (0, 1 / alpha)))

    @classmethod
    def gyrator(cls, R):

        R = ConstantDomainExpression(R)

        return cls(((0, R),
                    (1 / R, 0)))

    @classmethod
    def Lsection(cls, Z1, Z2):

        Y = 1 / Z2
        return cls(((1 + Y * Z1, -Z1),
                    (-Y, 1)))
        # return cls.Zseries(Z1).chain(cls.Zshunt(Z2))

    @classmethod
    def Tsection(cls, Z1, Z2, Z3):

        Y = 1 / Z2
        return cls(((1 + Y * Z1, -Z1 - Z3 * (1 + Y * Z1)),
                    (-Y, 1 + Y * Z3)))
        # return cls.Lsection(Z1, Z2).chain(cls.Zseries(Z3))

    @classmethod
    def Pisection(cls, Z1, Z2, Z3):

        return cls.Zshunt(Z1).chain(cls.Lsection(Z2, Z3))

    def chain(self, TP):

        # Note reverse order compared to AMatrix.
        return TP * self

    def cascade(self, TP):

        return self.chain(TP)


class GMatrix(TwoPortMatrix):

    """G-parameters (inverse hybrid parameters)

    ::
       +-  -+     +-       -+   +-  -+
       | V2 |  =  | G11  G12|   | I2 |
       | I1 |     | G21  G22|   | V1 |
       +-  -+     +-       -+   +-  -+

              +-         -+
       units  | ohm     1 |
              | 1   1/ohm |
              +-         -+

    G = inv(H)
    """

    @classmethod
    def generic(cls):
        return cls((('G_11', 'G_12'), ('G_21', 'G_22')))

    def equation(self):
        return Eq(Matrix(('I1', 'V2')), MatMul(self, Matrix(('V1', 'I2'))),
                  evaluate=False)
        
    @property
    def Aparams(self):
        # return self.Hparams.Aparams
        det = self.det().expr        
        return AMatrix(((1 / self._G21, self._G22 / self._G21),
                        (self._G11 / self._G21, det / self._G21)))

    @property
    def Bparams(self):
        # return self.Hparams.Bparams
        det = self.det().expr        
        return BMatrix(((-det / self._G12),
                        (self._G22 / self._G12, self._G11 / self._G12, -1 / self._G12)))

    @property
    def Gparams(self):
        # Perhaps we should make a copy?
        return self

    @property
    def Hparams(self):
        return HMatrix(self.inv())

    @property
    def Yparams(self):
        return self.Hparams.Yparams

    @property
    def Zparams(self):
        return self.Hparams.Zparams


class HMatrix(TwoPortMatrix):

    """H-parameters (hybrid parameters)
    ::
       +-  -+     +-       -+   +-  -+
       | V1 |  =  | H11  H12|   | I1 |
       | I2 |     | H21  H22|   | V2 |
       +-  -+     +-       -+   +-  -+

              +-         -+
       units  | ohm     1 |
              | 1   1/ohm |
              +-         -+

    H = inv(G)
    """

    @classmethod
    def generic(cls):
        return cls((('H_11', 'H_12'), ('H_21', 'H_22')))    

    def equation(self):
        return Eq(Matrix(('V1', 'I2')), MatMul(self, Matrix(('I1', 'V2'))),
                  evaluate=False)
        
    @property
    def Aparams(self):
        det = self.det().expr
        return AMatrix(((-det / self._H21, -self._H11 / self._H21),
                        (-self._H22 / self._H21, -1 / self._H21)))

    @property
    def Bparams(self):
        det = self.det().expr
        return BMatrix(((1 / self._H12, -self._H11 / self._H12),
                        (-self._H22 / self._H12, det / self._H12)))

    @property
    def Hparams(self):
        # Perhaps we should make a copy?
        return self

    @property
    def Yparams(self):
        det = self.det().expr        
        return YMatrix(((1 / self._H11, -self._H12 / self._H11),
                        (self._H21 / self._H11, det / self._H11)))

    @property
    def Zparams(self):
        det = self.det().expr        
        return ZMatrix(((det / self._H22, self._H12 / self._H22),
                        (-self._H21 / self._H22, 1 / self._H22)))


class SMatrix(TwoPortMatrix):

    """S-parameters (scattering parameters)
    ::
       +-  -+     +-       -+   +-  -+
       | b1 |  =  | S11  S12|   | a1 |
       | b2 |     | S21  S22|   | a2 |
       +-  -+     +-       -+   +-  -+    

    Each element in the s-matrix has units of impedance.
    """    

    @classmethod
    def generic(cls):
        return cls((('S_11', 'S_12'), ('S_21', 'S_22')))

    def equation(self):
        return Eq(Matrix(('b1', 'b2')), MatMul(self, Matrix(('a1', 'a2'))),
                  evaluate=False)

    @property
    def Aparams(self):
        det = self.det().expr
        Z0 = LaplaceDomainImpedance('Z_0').as_expr()
        A = AMatrix(((1 + self._S11 - self._S22 - det,
                      (1 + self._S11 + self._S22 + det) * Z0),
                     ((1 - self._S11 - self._S22 + det) / Z0,
                      1 - self._S11 + self._S22 - det))) / (2 * self._S21)
        return A.simplify()

    @property
    def Hparams(self):
        return self.Aparams.Hparams

    @property
    def Sparams(self):
        return self

    @property
    def Tparams(self):
        det = self.det().expr        
        return TMatrix(((-det / self._S21, self._S11 / self._S21),
                        (-self._S22 / self._S21, 1 / self._S21))).simplify()

    @property
    def Zparams(self):
        return self.Aparams.Zparams
        
    
class TMatrix(TwoPortMatrix):
    
    """T-parameters (scattering transfer parameters)
    ::
       +-  -+     +-       -+   +-  -+
       | b1 |  =  | T11  T12|   | a2 |
       | a2 |     | T21  T22|   | b2 |
       +-  -+     +-       -+   +-  -+    
    """

    @classmethod
    def generic(cls):
        return cls((('T_11', 'T_12'), ('T_21', 'T_22')))
    
    # Note, another convention uses a1, b1 in terms of b2, a2.
    def equation(self):
        return Eq(Matrix(('b1', 'a1')), MatMul(self, Matrix(('a2', 'b2'))),
                  evaluate=False)

    @property
    def Aparams(self):
        return self.Sparams.Aparams
    
    @property
    def Hparams(self):
        return self.Aparams.Hparams

    @property
    def Sparams(self):
        det = self.det().expr        
        return SMatrix(((self._T12 / self._T22, det / self._T22),
                        (1 / self._T22, -self._T21 / self._T22)))
    
    @property
    def Tparams(self):
        return self    

    @property
    def Zparams(self):
        return self.Aparams.Zparams


    
class YMatrix(TwoPortMatrix):

    """Y-parameters (admittance parameters)
    ::
       +-  -+     +-       -+   +-  -+
       | I1 |  =  | Y11  Y12|   | V1 |
       | I2 |     | Y21  Y22|   | V2 |
       +-  -+     +-       -+   +-  -+

              +-           -+
       units  | 1/ohm 1/ohm |
              | 1/ohm 1/ohm |
              +-           -+

    Y = inv(Z)
    """

    @classmethod
    def generic(cls):
        return cls((('Y_11', 'Y_12'), ('Y_21', 'Y_22')))

    def equation(self):
        return Eq(Matrix(('I1', 'I2')), MatMul(self, Matrix(('V1', 'V2'))),
                  evaluate=False)
        
    @property
    def Ysc(self):
        return LaplaceDomainAdmittanceMatrix((self._Y11, self._Y22))

    @property
    def Aparams(self):
        det = self.det().expr        
        return AMatrix(((-self._Y22 / self._Y21, -1 / self._Y21),
                        (-det / self._Y21, -self._Y11 / self._Y21)))

    @property
    def Bparams(self):
        det = self.det().expr        
        return BMatrix(((-self._Y11 / self._Y12, 1 / self._Y12),
                        (det / self._Y12, -self._Y22 / self._Y12)))

    @property
    def Hparams(self):
        det = self.det().expr
        return HMatrix(((1 / self._Y11, -self._Y12 / self._Y11),
                        (self._Y21 / self._Y11, det / self._Y11)))

    @property
    def Yparams(self):
        # Perhaps we should make a copy?
        return self

    @property
    def Zparams(self):
        # Inverse
        det = self.det().expr
        return ZMatrix(((self._Y22 / det, -self._Y12 / det),
                        (-self._Y21 / det, self._Y11 / det)))


class ZMatrix(TwoPortMatrix):

    """Z-parameters (impedance parameters)
    ::
       +-  -+     +-       -+   +-  -+
       | V1 |  =  | Z11  Z12|   | I1 |
       | V2 |     | Z21  Z22|   | I2 |
       +-  -+     +-       -+   +-  -+

              +-         -+
       units  | ohm   ohm |
              | ohm   ohm |
              +-         -+

    Z = inv(Y)
    """

    @classmethod
    def generic(cls):
        return cls((('Z_11', 'Z_12'), ('Z_21', 'Z_22')))

    def equation(self):
        return Eq(Matrix(('V1', 'V2')), MatMul(self, Matrix(('I1', 'I2'))),
                  evaluate=False)
        
    @property
    def Zoc(self):
        return LaplaceDomainImpedanceMatrix((self._Z11, self._Z22))

    @property
    def Aparams(self):
        det = self.det().expr
        return AMatrix(((self._Z11 / self._Z21, det / self._Z21),
                        (1 / self._Z21, self._Z22 / self._Z21)))

    @property
    def Bparams(self):
        det = self.det().expr
        return BMatrix(((self._Z22 / self._Z12, -det / self._Z12),
                        (-1 / self._Z12, self._Z11 / self._Z12)))

    @property
    def Hparams(self):
        det = self.det().expr
        return HMatrix(((det / self._Z22, self._Z12 / self._Z22),
                        (-self._Z21 / self._Z22, 1 / self._Z22)))

    @property
    def Yparams(self):
        # Inverse
        det = self.det().expr
        return YMatrix(((self._Z22 / det, -self._Z12 / det),
                        (-self._Z21 / det, self._Z11 / det)))

    @property
    def Zparams(self):
        # Perhaps we should make a copy?
        return self

    @classmethod
    def Lsection(cls, Z1, Z2):
        return cls.Tsection(Z1 + Z2, Z2, Z2, Z2)

    @classmethod
    def Tsection(cls, Z1, Z2, Z3):
        # Note if Z3 is infinity then all elements of Z are infinite.
        # Thus we cannot model a single series R with a Z matrix.
        # A single shunt R works though.
        return cls(((Z1 + Z2, Z2),
                    (Z2, Z2 + Z3)))

    @classmethod
    def Pisection(cls, Z1, Z2, Z3):

        Za, Zb, Zc = DeltaWye(Z1, Z2, Z3)
        return cls.Tsection(Za, Zb, Zc)


class TwoPort(Network, TwoPortMixin):

    """
    General class for two-port networks.  Two-port networks are
    constrained to have the same current at each port (but flowing in
    opposite directions).  This is called the port condition.
    """

    def _add_elements(self):
        raise ValueError('Cannot generate netlist for two-port objects')

    def netlist(self):
        raise ValueError('Cannot generate netlist for two-port objects')

    def _check_twoport_args(self):

        # This is an interim measure until Par2, Ser2, etc. generalised.
        if len(self.args) != 2:
            raise ValueError('Only two args supported for %s' %
                             self.__class__.__name__)
        for arg1 in self.args:
            if not isinstance(arg1, TwoPort):
                raise ValueError('%s not a TwoPort' % arg1)

    @property
    def Aparams(self):
        """Return chain parameters"""
        return self._M.Aparams

    @property
    def Bparams(self):
        """Return inverse chain parameters"""
        return self._M.Bparams

    @property
    def Gparams(self):
        """Return inverse hybrid parameters"""
        return self._M.Gparams

    @property
    def Hparams(self):
        """Return hybrid parameters"""
        return self._M.Hparams

    @property
    def Sparams(self):
        """Return scattering parameters"""
        return self._M.Sparams

    @property
    def Tparams(self):
        """Return scattering transfer parameters"""
        return self._M.Tparams    

    @property
    def Yparams(self):
        """Return admittance parameters"""
        return self._M.Yparams

    @property
    def Zparams(self):
        """Return impedance parameters"""
        return self._M.Zparams

    @property
    def I1a(self):
        return LaplaceDomainCurrent(-self.V2b / self._B12)

    @property
    def V1a(self):
        # CHECKME
        return LaplaceDomainVoltage(-self.I2b / self._B21)

    @property
    def I1g(self):
        return LaplaceDomainCurrent(-self.I2b / self._B22)

    @property
    def V2g(self):
        return LaplaceDomainVoltage(self.V2b - self._B12 / self._B22 * self.I2b)

    @property
    def V1h(self):
        return LaplaceDomainVoltage(-self.V2b / self._B11)

    @property
    def I2h(self):
        return LaplaceDomainCurrent(-self.V2b * self._B21 / self._B11) - self.I2b

    @property
    def I1y(self):
        return LaplaceDomainCurrent(-self.V2b / self._B12)

    @property
    def I2y(self):
        return LaplaceDomainCurrent(self.V2b * self._B22 / self._B12) - self.I2b

    @property
    def V1z(self):
        return LaplaceDomainVoltage(-self.I2b / self._B21)

    @property
    def V2z(self):
        return self.V2b - LaplaceDomainVoltage(self.I2b * self._B11 / self._B21)

    @property
    def Yoc(self):
        """Return admittance vector with ports open circuit"""
        return LaplaceDomainAdmittanceMatrix((LaplaceDomainAdmittance(1 / self.Z1oc),
                                              LaplaceDomainAdmittance(1 / self.Z2oc)))

    @property
    def Y1oc(self):
        """Return input impedance with the output port open circuit"""
        return LaplaceDomainImpedance(1 / self.Z1oc)

    @property
    def Y2oc(self):
        """Return output impedance with the input port open circuit"""
        return LaplaceDomainAdmittance(1 / self.Z2oc)

    @property
    def Ysc(self):
        """Return admittance vector with ports short circuit"""
        return self.Yparams.Ysc

    @property
    def Y1sc(self):
        """Return input admittance with output port short circuit"""
        return LaplaceDomainAdmittance(self.Ysc[0])

    @property
    def Y2sc(self):
        """Return output admittance with output port short circuit"""
        return LaplaceDomainAdmittance(self.Ysc[1])

    @property
    def Zoc(self):
        """Return impedance vector with ports open circuit"""
        return self.Zparams.Zoc

    @property
    def Z1oc(self):
        """Return input impedance with the output port open circuit"""
        return LaplaceDomainImpedance(self.Zoc[0])

    @property
    def Z2oc(self):
        """Return output impedance with the input port open circuit"""
        return LaplaceDomainImpedance(self.Zoc[1])

    @property
    def Zsc(self):
        """Return impedance vector with ports short circuit"""
        return LaplaceDomainImpedanceMatrix((LaplaceDomainImpedance(1 / self.Y1sc),
                                             LaplaceDomainImpedance(1 / self.Y2sc)))

    @property
    def Z1sc(self):
        """Return input impedance with the output port short circuit"""
        return LaplaceDomainImpedance(1 / self.Y1sc)

    @property
    def Z2sc(self):
        """Return output impedance with the input port short circuit"""
        return LaplaceDomainImpedance(1 / self.Y2sc)

    def Vgain(self, inport=1, outport=2):
        """Return voltage gain for specified ports with internal
        sources zero"""

        # Av  = G21 = 1 / A11 = -det(B) / B22 = Z21 / Z11 =  Y21 / Y22
        # Av' = H12 = 1 / B11 =  |A| / A22 = Z12 / Z22 = -Y12 / Y11

        if inport == outport:
            return LaplaceDomainTransferFunction(1)
        if inport == 1 and outport == 2:
            return LaplaceDomainTransferFunction(1 / self._A11)
        if inport == 2 and outport == 1:
            return LaplaceDomainTransferFunction(1 / self._B11)
        raise ValueError('bad port values')

    def Igain(self, inport=1, outport=2):
        """Return current gain for specified ports with internal
         sources zero"""

        # Ai  = H21 = -1 / A22 = -det(B) / B11 = -Z21 / Z22 = Y21 / Y11
        # Ai' = G12 =  1 / B22 =  |A| / A11 = -Z12 / Z11 = Y12 / Y22

        if inport == outport:
            return LaplaceDomainTransferFunction(1)
        if inport == 1 and outport == 2:
            return LaplaceDomainTransferFunction(-1 / self._A22)
        if inport == 2 and outport == 1:
            return LaplaceDomainTransferFunction(-1 / self._B22)
        raise ValueError('bad port values')

    @property
    def Vgain12(self):
        """Return V2 / V1 for I2 = 0 (forward voltage gain) with
        internal sources zero

        Av = G21 = 1 / A11 = -det(B) / B22 = Z21 / Z11 =  Y21 / Y22
        """

        return self.Vgain(1, 2)

    @property
    def Vtransfer(self):
        """Return V2 / V1 for I2 = 0 (forward voltage gain) with
        internal sources zero  (see Vgain12)"""

        return self.Vgain12

    @property
    def Igain12(self):
        """Return I2 / I1 for V2 = 0 (forward current gain) with
        internal sources zero

        Ai = H21 = -1 / A22 = -det(B) / B11 = -Z21 / Z22 = Y21 / Y11
        """

        return self.Igain(1, 2)

    @property
    def Itransfer(self):
        """Return I2 / I1 for V2 = 0 (forward current gain) with
        internal sources zero  (sett Igain12)"""

        return self.Igain12

    def Vresponse(self, V, inport=1, outport=2):
        """Return voltage response for specified applied voltage and
        specified ports"""

        if issubclass(V.__class__, OnePort):
            V = V.Voc.laplace()

        p1 = inport - 1
        p2 = outport - 1

        H = self.Zparams[p2, p1] / self.Zparams[p1, p1]
        return LaplaceDomainVoltage(self.Voc[p2] + (V - self.Voc[p1]) * H)

    def Iresponse(self, I, inport=1, outport=2):
        """Return current response for specified applied current and
        specified ports"""

        if issubclass(I.__class__, OnePort):
            I = I.Isc.laplace()

        p1 = inport - 1
        p2 = outport - 1

        Y = self.Yparams
        Isc = self.Isc

        return LaplaceDomainCurrent(Isc[p2] + Y[p2, p1] / Y[p1, p1] * (I - Isc[p1]))

    def Ytrans(self, inport=1, outport=2):
        """Return transadmittance for specified ports with internal
        sources zero"""

        return LaplaceDomainAdmittance(self.Yparams[outport - 1, inport - 1])

    @property
    def Ytrans12(self):
        """Return I2 / V1 for V2 = 0 (forward transadmittance) with
         internal sources zero

         Y21 = -1 / A12 = det(B) / B12
         """

        return LaplaceDomainAdmittance(self._Y21)

    @property
    def Ytransfer(self):
        """Return I2 / V1 for V2 = 0 (forward transadmittance) with
         internal sources zero.  This is an alias for Ytrans12.

         Y21 = -1 / A12 = det(B) / B12
         """

        return self.Ytrans12

    def Ztrans(self, inport=1, outport=2):
        """Return transimpedance for specified ports with internal
        sources zero"""

        return LaplaceDomainImpedance(self.Zparams[outport - 1, inport - 1])

    def Ztrans12(self):
        """Return V2 / I1 for I2 = 0 (forward transimpedance) with
        internal sources zero

        Z21 = 1 / A21 = -det(B) / B21
        """

        return LaplaceDomainImpedance(self._Z21)

    @property
    def Ztransfer(self):
        """Return V2 / I1 for I2 = 0 (forward transimpedance) with
        internal sources zero.  This is an alias for Ztrans12.

        Z21 = 1 / A21 = -det(B) / B21
        """

        return self.Ztrans12

    @property
    def V1oc(self):
        """Return V1 with all ports open-circuited (i.e., I1 = I2 = 0)"""
        return LaplaceDomainVoltage(self.Voc[0])

    @property
    def V2oc(self):
        """Return V2 with all ports open-circuited (i.e., I1 = I2 = 0)"""
        return LaplaceDomainVoltage(self.Voc[1])

    @property
    def I1sc(self):
        """Return I1 with all ports short-circuited, i.e, V1 = V2 = 0"""
        return LaplaceDomainCurrent(self.Isc[0])

    @property
    def I2sc(self):
        """Return I2 with all ports short-circuited, i.e, V1 = V2 = 0"""
        return LaplaceDomainCurrent(self.Isc[1])

    @property
    def Voc(self):
        """Return voltage vector with all ports open-circuited
        (i.e., In = 0)"""
        return LaplaceDomainVoltageMatrix((self.V1z, self.V2z))

    @property
    def Isc(self):
        """Return current vector with all ports short-circuited
        (i.e., V1 = V2 = 0)"""
        return LaplaceDomainCurrentMatrix((self.I1y, self.I2y))

    @property
    def Bmodel(self):

        return TwoPortBModel(self.Bparams, self.V2b, self.I2b)

    @property
    def Hmodel(self):

        return TwoPortHModel(self.Hparams, self.V1h, self.I2h)

    @property
    def Ymodel(self):

        if self.is_shunt:
            warn('Converting a shunt two-port to a Y model is dodgy...')
        return TwoPortYModel(self.Yparams, self.I1y, self.I2y)

    @property
    def Zmodel(self):

        if self.is_series:
            warn('Converting a series two-port to a Z model is dodgy...')
        return TwoPortZModel(self.Zparams, self.V1z, self.V2z)

    def chain(self, TP):
        """Return the model with, TP, appended (cascade or
        chain connection)"""

        if not issubclass(TP.__class__, TwoPort):
            raise TypeError('Argument not', TwoPort)

        return Chain(self, TP)

    def append(self, TP):
        """Return the model with, TP, appended"""

        return self.chain(TP)

    def prepend(self, TP):
        """Return the model with, TP, prepended"""

        return TP.chain(self)

    def cascade(self, TP):
        """Return the model with, TP, appended"""

        return self.chain(TP)

    def series(self, TP, port=None):
        """Return the model with, TP, in series.

         In general, this is tricky to ensure that the port condition
         is valid.  The common ground connection of the first two-port
         shorts out the top of the T of the second two-port.
         """

        if issubclass(TP.__class__, OnePort):
            raise NotImplementedError('TODO')

        warn('Do you mean chain?  The result of a series combination'
             ' of two two-ports may be dodgy')

        return Ser2(self, TP)

    def terminate(self, OP, port=2):
        """Connect one-port in parallel to specified port and return
        a Thevenin (one-port) object"""

        if port == 1:
            return self.source(OP)
        if port == 2:
            return self.load(OP)
        raise ValueError('Invalid port ' + port)

    def parallel(self, TP, port=None):
        """Return the model with, TP, in parallel"""

        if issubclass(TP.__class__, OnePort):
            raise NotImplementedError('TODO')

        return Par2(self, TP)

    def hybrid(self, TP, port=None):
        """Return the model with, TP, in hybrid connection (series
        input, parallel output)"""

        if issubclass(TP.__class__, OnePort):
            raise NotImplementedError('TODO')

        return Hybrid2(self, TP)

    def inverse_hybrid(self, TP, port=None):
        """Return the model with, TP, in inverse hybrid connection
        (parallel input, series output)"""

        if issubclass(TP.__class__, OnePort):
            raise NotImplementedError('TODO')

        return InverseHybrid2(self, TP)

    # Other operations: swapping the input terminals negates the A matrix.
    # switching ports.

    def bridge(self, TP):
        """Bridge the ports with a one-port element"""

        if not issubclass(TP.__class__, OnePort):
            raise TypeError('Argument not ', OnePort)

        # FIXME
        return self.parallel(Series(TP))

    def load(self, TP):
        """Apply a one-port load and return a Thevenin (one-port) object"""

        if not issubclass(TP.__class__, OnePort):
            raise TypeError('Argument not ', OnePort)

        foo = self.chain(Shunt(TP))
        return Z(foo.Z1oc) + V(foo.V1oc)

    def source(self, TP):
        """Apply a one-port source and return a Thevenin (one-port) object"""

        if not issubclass(TP.__class__, OnePort):
            raise TypeError('Argument not ', OnePort)

        foo = Shunt(TP).chain(self)
        return Z(foo.Z2oc) +  V(foo.V2oc)

    def short_circuit(self, port=2):
        """Apply a short-circuit to specified port and return a
        one-port object"""

        p = port - 1
        Yval = self.Yparams[1 - p, 1 - p]
        Ival = self.Isc[1 - p]

        return Y(Yval) | I(Ival)

    def open_circuit(self, port=2):
        """Apply a open-circuit to specified port and return a
        one-port object"""

        p = port - 1
        Zval = self.Zparams[1 - p, 1 - p]
        Vval = self.Voc[1 - p]

        return Z(Zval) + V(Vval)

    def simplify(self):

        if self.Bparams == sym.eye(2):
            # Have a pair of wires... perhaps could simplify
            # to an LSection comprised of a V and I but
            # may have a weird voltage expression.
            pass
        return self


class TwoPortBModel(TwoPort):

    """
    ::
            +-------------------+    +------+
     I1     |                   | I2'| -  + |          I2
    -->-----+                   +-<--+  V2b +----+-----<--
            |    two-port       |    |      |    |
    +       |    network        | +  +------+    |       +
    V1      |    without        | V2'        +---+---+  V2
    -       |    sources        | -          |   |   |   -
            |    represented    |               I2b  |
            |    by B matrix    |            |   v   |
            |                   |            +---+---+
            |                   |                |
    --------+                   +----------------+--------
            |                   |
            +-------------------+

    +-   +     +-        -+   +-  -+     +-   -+
    | V2 |  =  | B11  B12 |   | V1 |  +  | V2b |
    |-I2 |     | B21  B22 |   | I1 |     | I2b |
    +-  -+     +-        -+   +-  -+     +-   -+


    +-    +     +-        -+   +-  -+
    | V2' |  =  | B11  B12 |   | V1 |
    |-I2' |     | B21  B22 |   | I1 |
    +-   -+     +-        -+   +-  -+

    +-    +     +-  -+    +-   -+
    | V2' |  =  | V2 | -  | V2b |
    | I2' |     | I2'|    | I2b |
    +-  - +     +-  -+    +-   -+

    +-         +     +-        -+   +-  -+
    | V2 - V2b |  =  | B11  B12 |   | V1 |
    |-I2 - I2b |     | B21  B22 |   | I1 |
    +-        -+     +-        -+   +-  -+

    +-   +     +-        -+   +-    +
    | V1 |  =  | A11  A12 |   | V2' |
    | I1 |     | A21  A22 |   |-I2' |
    +-  -+     +-        -+   +-   -+

    +-   +     +-        -+   +-       -+
    | V1 |  =  | A11  A12 |   | V2 - V2b |
    | I1 |     | A21  A22 |   |-I2 + I2b |
    +-  -+     +-        -+   +-        -+

    """

    # A disadvantage of the Y and Z matrices is that they become
    # singular for some simple networks.  For example, the Z matrix is
    # singular for a shunt element and the Y matrix is singular for a
    # series element.  The A and B matrices do not seem to have this
    # problem, however, they cannot be extended to three or more ports.
    #

    def __init__(self, B, V2b=None, I2b=None):

        if V2b is None:
            V2b = LaplaceDomainVoltage(0)
        if I2b is None:
            I2b = LaplaceDomainCurrent(0)            

        if issubclass(B.__class__, TwoPortBModel):
            B, V2b, I2b = B._M, B._V2b, B._I2b

        if not isinstance(B, BMatrix):
            raise ValueError('B not BMatrix')

        if not isinstance(V2b, LaplaceDomainVoltage):
            raise ValueError('V2b not LaplaceDomainVoltage')

        if not isinstance(I2b, LaplaceDomainCurrent):
            raise ValueError('I2b not LaplaceDomainCurrent')

        super(TwoPortBModel, self).__init__()
        self._M = B
        self._V2b = V2b
        self._I2b = I2b

    @property
    def Bparams(self):
        """Return chain matrix"""
        return self._M

    @property
    def I2b(self):
        return self._I2b

    @property
    def V2b(self):
        return self._V2b

    @property
    def V1h(self):
        return LaplaceDomainVoltage(-self.V2b / self._B11)

    @property
    def I2h(self):
        return LaplaceDomainCurrent(-self.V2b * self._B21 / self._B11) - self.I2b

    @property
    def I1y(self):
        return LaplaceDomainCurrent(-self.V2b / self._B12)

    @property
    def I2y(self):
        return LaplaceDomainCurrent(self.V2b * self._B22 / self._B12) - self.I2b

    @property
    def V1z(self):
        return LaplaceDomainVoltage(-self.I2b / self._B21)

    @property
    def V2z(self):
        return self.V2b - LaplaceDomainVoltage(self.I2b * self._B11 / self._B21)


class TwoPortGModel(TwoPort):

    """
    """

    def __init__(self, G, I1g=None, V2g=None):

        if I1g is None:
            I1g = LaplaceDomainCurrent(0)
        if V2g is None:
            V2g = LaplaceDomainVoltage(0)            
        
        if issubclass(G.__class__, TwoPortGModel):
            G, I1g, V2g = G._M, G._I1g, G._V2g

        if not isinstance(G, GMatrix):
            raise ValueError('G not GMatrix')

        if not isinstance(I1g, LaplaceDomainCurrent):
            raise ValueError('I1g not LaplaceDomainCurrent')

        if not isinstance(V2g, LaplaceDomainVoltage):
            raise ValueError('V2g not LaplaceDomainVoltage')

        super(TwoPortGModel, self).__init__()
        self._M = G
        self._V1g = I1g
        self._I2g = V2g

    @property
    def Gparams(self):
        """Return hybrid matrix"""
        return self._M

    @property
    def V2b(self):
        """Return V2b"""

        # FIXME
        return LaplaceDomainVoltage(self.I1g / self.Gparams._G12)

    @property
    def I2b(self):
        """Return I2b"""

        # FIXME
        return LaplaceDomainCurrent(self.G._G22 / self.G._G12 * self.I1g) - self.V2g

    @property
    def I1g(self):
        return self._V1g

    @property
    def V2g(self):
        return self._I2g


class TwoPortHModel(TwoPort):

    """
    ::
         +------+   +-------------------+
     I1  | +  - |   |                   | I2'          I2
    -->--+  V1h +---+                   +-<-------+-----<--
         |      |   |    two-port       |         |
    +    +------+   |    network        | +       |       +
    V1              |    without        | V2' ---+---+  V2
    -               |    sources        | -  |   |   |   -
                    |    represented    |    |  I2h  |
                    |    by H matrix    |    |   v   |
                    |                   |    +---+---+
                    |                   |        |
    ----------------+                   +--------+--------
                    |                   |
                    +-------------------+


    +-   +     +-        -+   +-  -+     +-   -+
    | V1 |  =  | H11  H12 |   | I1 |  +  | V1h |
    | I2 |     | H21  H22 |   | V2 |     | I2h |
    +-  -+     +-        -+   +-  -+     +-   -+
    """

    def __init__(self, H, V1h=None, I2h=None):

        if V1h is None:
            V1h = LaplaceDomainVoltage(0)
        if I2h is None:
            I2h = LaplaceDomainCurrent(0)            
        
        if issubclass(H.__class__, TwoPortHModel):
            H, V1h, I2h = H._M, H._V1h, H._I2h

        if not isinstance(H, HMatrix):
            raise ValueError('H not HMatrix')

        if not isinstance(V1h, LaplaceDomainVoltage):
            raise ValueError('V1h not LaplaceDomainVoltage')

        if not isinstance(I2h, LaplaceDomainCurrent):
            raise ValueError('I2h not LaplaceDomainCurrent')

        super(TwoPortHModel, self).__init__()
        self._M = H
        self._V1h = V1h
        self._I2h = I2h

    @property
    def Hparams(self):
        """Return hybrid matrix"""
        return self._M

    @property
    def V2b(self):
        """Return V2b"""

        return LaplaceDomainVoltage(self.V1h / self.Hparams._H12)

    @property
    def I2b(self):
        """Return I2b"""

        return LaplaceDomainCurrent(self.H._H22 / self.H._H12 * self.V1h) - self.I2h

    @property
    def V1h(self):
        return self._V1h

    @property
    def I2h(self):
        return self._I2h


class TwoPortYModel(TwoPort):

    """
    ::
                     +-------------------+
     I1              |                   | I2'           I2
    -->----+---------+                   +-<-------+-----<--
           |         |    two-port       |         |
    +      |       + |    network        | +       +       +
    V1 +---+---+  V1'|    without        | V2' +---+---+  V2
    -  |   |   |   - |    sources        | -   |   |   |   -
       |  I1y  |     |    represented    |     |  I2y  |
       |   v   |     |    by Y matrix    |     |   v   |
       +---+---+     |                   |     +---+---+
           |         |                   |         |
    -------+---------+                   +---------+--------
                     |                   |
                     +-------------------+

    +-   +     +-        -+   +-  -+     +-   -+
    | I1 |  =  | Y11  Y12 |   | V1 |  +  | I1y |
    | I2 |     | Y21  Y22 |   | V2 |     | I2y |
    +-  -+     +-        -+   +-  -+     +-   -+

    Ymn = Im / Vn for Vm = 0
    """

    def __init__(self, Y, I1y=None, I2y=None):

        if I1y is None:
            I1y = LaplaceDomainCurrent(0)
        if I2y is None:
            I2y = LaplaceDomainCurrent(0)            
        
        if issubclass(Y.__class__, TwoPortYModel):
            Y, I1y, I2y = Y._M, Y._I1y, Y._I2y

        if not isinstance(Y, YMatrix):
            raise ValueError('Y not YMatrix')

        if not isinstance(I1y, LaplaceDomainCurrent):
            raise ValueError('I1y not LaplaceDomainCurrent')
        if not isinstance(I2y, LaplaceDomainCurrent):
            raise ValueError('I2y not LaplaceDomainCurrent')

        super(TwoPortYModel, self).__init__()
        self._M = Y
        self._I1y = I1y
        self._I2y = I2y

    @property
    def Yparams(self):
        """Return admittance matrix"""
        return self._M

    @property
    def I2b(self):
        return LaplaceDomainCurrent(-self.I1y * self._Y11 * self._Y22 / self._Y12) - self.I2y

    @property
    def V2b(self):
        return LaplaceDomainVoltage(self.I1y * self._Y11 / self._Y22)

    @property
    def I1y(self):
        return self._I1y

    @property
    def I2y(self):
        return self._I2y


class TwoPortZModel(TwoPort):

    """
    ::
         +------+    +-------------------+    +------+
    I1   | +  - | I1'|                   | I2'| -  + |  I2
    -->--+  V1z +-->-+                   +-<--+  V2z +--<--
         |      |    |    two-port       |    |      |
    +    +------+  + |    network        | +  +------+    +
    V1            V1'|    without        | V2'           V2
    -              - |    sources        | -              -
                     |    represented    |
                     |    by Z matrix    |
                     |                   |
                     |                   |
    -----------------+                   +-----------------
                     |                   |
                     +-------------------+

    +-   +     +-        -+   +-  -+     +-   -+
    | V1 |  =  | Z11  Z12 |   | I1 |  +  | V1z |
    | V2 |     | Z21  Z22 |   | I2 |     | V2z |
    +-  -+     +-        -+   +-  -+     +-   -+

    """

    def __init__(self, Z, V1z=None, V2z=None):

        if V1z is None:
            V1z = LaplaceDomainVoltage(0)
        if V2z is None:
            V2z = LaplaceDomainVoltage(0)            
        
        if issubclass(Z.__class__, TwoPortZModel):
            Z, V1z, V2z = Z._M, Z._V1z, Z._V2z

        if not isinstance(Z, ZMatrix):
            raise ValueError('Z not ZMatrix')

        if not isinstance(V1z, LaplaceDomainVoltage):
            raise ValueError('V1z not LaplaceDomainVoltage')
        if not isinstance(V2z, LaplaceDomainVoltage):
            raise ValueError('V2z not LaplaceDomainVoltage')

        super(TwoPortZModel, self).__init__()
        self._M = Z
        self._V1z = V1z
        self._V2z = V2z

    @property
    def Zparams(self):
        """Return impedance matrix"""
        return self._M

    @property
    def I2b(self):
        return LaplaceDomainCurrent(self.V1z / self._Z12)

    @property
    def V2b(self):
        return self.V2z - LaplaceDomainVoltage(self.V1z * self._Z22 / self._Z12)

    @property
    def I1y(self):

        Zdet = self.Zparams.det().expr
        return LaplaceDomainCurrent(-self.V1z * self._Z22 / Zdet - self.V2z * self._Z12 / Zdet)

    @property
    def I2y(self):

        Zdet = self.Zparams.det().expr
        return LaplaceDomainCurrent(self.V1z * self._Z21 / Zdet - self.V2z * self._Z11 / Zdet)

    @property
    def V1z(self):
        return self._V1z

    @property
    def V2z(self):
        return self._V2z


class Chain(TwoPortBModel):

    """Connect two-port networks in a chain (aka cascade)"""

    def __init__(self, *args):

        self.args = args
        self._check_twoport_args()

        arg1 = args[-1]
        B = arg1.Bparams

        foo = Vector(arg1.V2b, arg1.I2b)

        for arg in reversed(args[0:-1]):

            foo += B * Vector(arg.V2b, arg.I2b)
            B = B * arg.Bparams

        super(Chain, self).__init__(B, LaplaceDomainVoltage(foo[0, 0]), LaplaceDomainCurrent(foo[1, 0]))

    def simplify(self):

        if isinstance(self.args[0], Shunt) and isinstance(self.args[1], Shunt):
            return Shunt(
                (self.args[0].args[0] | self.args[1].args[0]).simplify())

        if isinstance(self.args[0], Series) and isinstance(
                self.args[1], Series):
            return Series(
                (self.args[0].args[0] + self.args[1].args[0]).simplify())

        return self


class Par2(TwoPortYModel):

    """Connect two-port networks in parallel"""

    def __init__(self, *args):

        self.args = args
        self._check_twoport_args()

        # This will fail with a Shunt as an argument since it does
        # not have a valid Y model.
        # We can special case this.
        if isinstance(args[0], Shunt) or isinstance(args[1], Shunt):
            print('Warning: need to handle a Shunt in parallel')

        arg = args[0]
        I1y = arg.I1y
        I2y = arg.I2y
        Y = arg.Yparams

        for arg in args[1:]:
            I1y += arg.I1y
            I2y += arg.I2y
            Y += arg.Yparams

        super(Par2, self).__init__(Y, I1y, I2y)

    def simplify(self):

        if isinstance(self.args[0], Shunt) and isinstance(self.args[1], Shunt):
            return Shunt(
                (self.args[0].args[0] | self.args[1].args[0]).simplify())

        if isinstance(self.args[0], Series) and isinstance(
                self.args[1], Series):
            return Series(
                (self.args[0].args[0] | self.args[1].args[0]).simplify())

        return self


class Ser2(TwoPortZModel):

    """Connect two-port networks in series (note this is unusual and can
    break the port condition)"""

    def __init__(self, *args):

        self.args = args
        self._check_twoport_args()

        # Need to be more rigorous.
        if isinstance(self.args[1], (Series, LSection, TSection)):
            print('Warning: This can violate the port condition')

        arg = args[0]
        V1z = arg.V1z
        V2z = arg.V2z
        Z = arg.Zparams

        for arg in args[1:]:
            V1z += arg.V1z
            V2z += arg.V2z
            Z += arg.Zparams

        super(Ser2, self).__init__(Z, V1z, V2z)

    def simplify(self):

        if isinstance(self.args[0], Shunt) and isinstance(self.args[1], Shunt):
            return Shunt(
                (self.args[0].args[0] + self.args[1].args[0]).simplify())

        return self


class Hybrid2(TwoPortHModel):

    """Connect two-port networks in hybrid configuration (inputs in
    series, outputs in parallel)"""

    def __init__(self, *args):

        self.args = args
        self._check_twoport_args()

        arg = args[0]
        V1h = arg.V1h
        I2h = arg.I2h
        H = arg.Hparams

        for arg in args[1:]:
            V1h += arg.V1h
            I2h += arg.I2h
            H += arg.Hparams

        super(Hybrid2, self).__init__(H, V1h, I2h)


class InverseHybrid2(TwoPortGModel):

    """Connect two-port networks in inverse hybrid configuration (outputs in
    series, inputs in parallel)"""

    def __init__(self, *args):

        self.args = args
        self._check_twoport_args()

        arg = args[0]
        I1g = arg.I1g
        V2g = arg.V2g
        G = arg.Gparams

        for arg in args[1:]:
            I1g += arg.I1g
            V2g += arg.V2g
            G += arg.Gparams

        super(Hybrid2, self).__init__(G, I1g, V2g)


class Series(TwoPortBModel):

    """
    Two-port comprising a single one-port in series configuration
    ::

           +---------+
         --+   OP    +---
           +---------+

         ----------------

    Note, this has a singular Y matrix.
    """

    def __init__(self, OP):

        self.OP = OP
        self.args = (OP, )
        _check_oneport_args(self.args)
        super(Series, self).__init__(BMatrix.Zseries(OP.Z.laplace()), LaplaceDomainVoltage(OP.Voc.laplace()), LaplaceDomainCurrent(0))


class Shunt(TwoPortBModel):

    """
    Two-port comprising a single one-port in shunt configuration
    ::

         -----+----
              |
            +-+-+
            |   |
            |OP |
            |   |
            +-+-+
              |
         -----+----

    Note, this has a singular Z matrix.
    """

    def __init__(self, OP):

        self.OP = OP
        self.args = (OP, )
        _check_oneport_args(self.args)
        super(Shunt, self).__init__(BMatrix.Yshunt(OP.Y.laplace()), LaplaceDomainVoltage(0),
                                    LaplaceDomainCurrent(OP.Isc.laplace()))


class IdealTransformer(TwoPortBModel):

    """Ideal transformer voltage gain alpha, current gain 1 / alpha"""

    def __init__(self, alpha=1):

        self.alpha = ConstantDomainExpression(alpha)
        self.args = (alpha, )
        super(IdealTransformer, self).__init__(BMatrix.transformer(alpha))


class TF(IdealTransformer):
    pass


class IdealGyrator(TwoPortBModel):

    """Ideal gyrator with gyration resistance R.

    A gyrator converts a voltage to current and a current to voltage.
    Cascaded gyrators act like a transformer"""

    def __init__(self, R=1):

        self.R = ConstantDomainExpression(R)
        self.args = (R, )
        super(IdealGyrator, self).__init__(BMatrix.gyrator(R))


class VoltageFollower(TwoPortBModel):

    """Voltage follower"""

    def __init__(self):

        self.args = ()
        super(VoltageFollower, self).__init__(BMatrix.voltage_amplifier(1))


class VoltageAmplifier(TwoPortBModel):

    """Voltage amplifier"""

    def __init__(self, Av=1, Af=0, Yin=0, Zout=0):

        Av = LaplaceDomainExpression(Av)
        Af = LaplaceDomainExpression(Af)
        Yin = LaplaceDomainExpression(Yin)
        Zout = LaplaceDomainExpression(Zout)

        self.args = (Av, Af, Yin, Zout)
        super(VoltageAmplifier, self).__init__(
            BMatrix.voltage_amplifier(Av, Af, Yin, Zout))


class IdealVoltageAmplifier(VoltageAmplifier):

    """Ideal voltage amplifier"""

    def __init__(self, Av=1):

        Av = LaplaceDomainExpression(Av)
        super(IdealVoltageAmplifier, self).__init__(
            BMatrix.voltage_differentiator(Av))
        self.args = (Av, )


class IdealDelay(TwoPortBModel):

    """Ideal buffered delay"""

    def __init__(self, delay=0):

        delay = ConstantDomainExpression(delay)
        super(IdealDelay, self).__init__(
            BMatrix.voltage_amplifier(sym.exp(-s * delay)))
        self.args = (delay, )


class IdealVoltageDifferentiator(TwoPortBModel):

    """Voltage differentiator"""

    def __init__(self, Av=1):

        Av = LaplaceDomainExpression(Av)
        super(IdealVoltageDifferentiator, self).__init__(
            BMatrix.voltage_differentiator(Av))
        self.args = (Av, )


class IdealVoltageIntegrator(TwoPortBModel):

    """Ideal voltage integrator"""

    def __init__(self, Av=1):

        Av = LaplaceDomainExpression(Av)
        super(IdealVoltageIntegrator, self).__init__(
            BMatrix.voltage_integrator(Av))
        self.args = (Av, )


class CurrentFollower(TwoPortBModel):

    """Current follower"""

    def __init__(self):

        super(CurrentFollower, self).__init__(BMatrix.current_amplifier(1))
        self.args = ()


class IdealCurrentAmplifier(TwoPortBModel):

    """Ideal current amplifier"""

    def __init__(self, Ai=1):

        Ai = LaplaceDomainExpression(Ai)
        super(IdealCurrentAmplifier, self).__init__(
            BMatrix.current_amplifier(Ai))
        self.args = (Ai, )


class IdealCurrentDifferentiator(TwoPortBModel):

    """Ideal current differentiator"""

    def __init__(self, Ai=1):

        Ai = LaplaceDomainExpression(Ai)
        super(IdealCurrentDifferentiator, self).__init__(
            BMatrix.current_differentiator(Ai))
        self.args = (Ai, )


class IdealCurrentIntegrator(TwoPortBModel):

    """Ideal current integrator"""

    def __init__(self, Ai=1):

        Ai = LaplaceDomainExpression(Ai)
        super(IdealCurrentIntegrator, self).__init__(
            BMatrix.current_integrator(Ai))
        self.args = (Ai, )


class OpampInverter(TwoPortBModel):

    """Opamp inverter"""

    def __init__(self, R1, R2):

        R1 = ConstantDomainExpression(R1)
        R2 = ConstantDomainExpression(R2)
        # FIXME for initial voltages.
        super(OpampInverter, self).__init__(
            AMatrix(((-R1.Z / R2.Z, 0),
                     (-1 / R2.Z, 0).Bparams)))
        self.args = (R1, R2)


class OpampIntegrator(TwoPortBModel):

    """Inverting opamp integrator"""

    def __init__(self, R1, C1):

        R1 = ConstantDomainExpression(R1)
        C1 = ConstantDomainExpression(C1)
        # FIXME for initial voltages.
        super(OpampIntegrator, self).__init__(
            AMatrix(((-R1.Z / C1.Z, 0), (-1 / C1.Z, 0))).Bparams)
        self.args = (R1, C1)


class OpampDifferentiator(TwoPortBModel):

    """Inverting opamp differentiator"""

    def __init__(self, R1, C1):

        R1 = ConstantDomainExpression(R1)
        C1 = ConstantDomainExpression(C1)
        # FIXME for initial voltages.
        super(OpampDifferentiator, self).__init__(
            AMatrix(((-R1.Z * C1.Z, 0), (-R1.Z, 0))).Bparams)
        self.args = (R1, C1)


class TSection(TwoPortBModel):

    """T (Y) section
    ::

           +---------+       +---------+
         --+   OP1   +---+---+   OP3   +---
           +---------+   |   +---------+
                       +-+-+
                       |   |
                       |OP2|
                       |   |
                       +-+-+
                         |
         ----------------+-----------------

      The Z matrix for a resistive T section is
      [ R1 + R2, R2     ]
      [      R2, R2 + R3]
    """

    def __init__(self, OP1, OP2, OP3):

        self.args = (OP1, OP2, OP3)
        _check_oneport_args(self.args)
        super(TSection, self).__init__(
            Series(OP1).chain(Shunt(OP2)).chain(Series(OP3)))

    def Pisection(self):

        ZV = WyeDelta(self.args[0].Z, self.args[1].Z, self.args[2].Z)
        VV = WyeDelta(self.args[0].V, self.args[1].V, self.args[2].V)
        OPV = [(ZV1 + VV1).cpt() for ZV1, VV1 in zip(ZV, VV)]

        return PiSection(*OPV)


class TwinTSection(TwoPortBModel):

    """Twin T section
    ::

              +---------+       +---------+
           +--+   OP1a  +---+---+   OP3a  +--+
           |  +---------+   |   +---------+  |
           |              +-+-+              |
           |              |   |              |
           |              |OP2a              |
           |              |   |              |
           |              +-+-+              |
           |                |                |
           |                v                |
           |  +---------+       +---------+  |
         --+--+   OP1b  +---+---+   OP3b  +--+--
              +---------+   |   +---------+
                          +-+-+
                          |   |
                          |OP2b
                          |   |
                          +-+-+
                            |
         -------------------+--------------------

    """

    def __init__(self, OP1a, OP2a, OP3a, OP1b, OP2b, OP3b):

        self.args = (OP1a, OP2a, OP3a, OP1b, OP2b, OP3b)
        _check_oneport_args(self.args)

        super(TwinTSection, self).__init__(
            TSection(OP1a, OP2a, OP3a).parallel(TSection(OP1b, OP2b, OP3b)))


class BridgedTSection(TwoPortBModel):

    """Bridged T section
        ::

                       +---------+
           +-----------+   OP4   +-----------+
           |           +---------+           |
           |                                 |
           |  +---------+       +---------+  |
         --+--+   OP1b  +---+---+   OP3b  +--+--
              +---------+   |   +---------+
                          +-+-+
                          |   |
                          |OP2b
                          |   |
                          +-+-+
                            |
         -------------------+--------------------

         """

    def __init__(self, OP1, OP2, OP3, OP4):

        self.args = (OP1, OP2, OP3, OP4)
        _check_oneport_args(self.args)

        super(TwinTSection, self).__init__(
            TSection(OP1, OP2, OP3).parallel(Series(OP4)))


class PiSection(TwoPortBModel):

    """Pi (delta) section
    ::

                  +---------+
        -----+----+   OP2    +---+-----
             |    +---------+   |
           +-+-+              +-+-+
           |   |              |   |
           |OP1|              |OP3|
           |   |              |   |
           +-+-+              +-+-+
             |                  |
        -----+------------------+-----

    """

    def __init__(self, OP1, OP2, OP3):

        super(PiSection, self).__init__(
            Shunt(OP1).chain(Series(OP2)).chain(Shunt(OP3)))
        self.args = (OP1, OP2, OP3)

    def Tsection(self):

        ZV = DeltaWye(self.args[0].Z, self.args[1].Z, self.args[2].Z)
        VV = DeltaWye(self.args[0].V, self.args[1].V, self.args[2].V)
        OPV = [(ZV1 + VV1).cpt() for ZV1, VV1 in zip(ZV, VV)]
        return TSection(*OPV)


class LSection(TwoPortBModel):

    """L Section
    ::

           +---------+
         --+   OP1   +---+----
           +---------+   |
                       +-+-+
                       |   |
                       |OP2|
                       |   |
                       +-+-+
                         |
         ----------------+----
    """

    def __init__(self, OP1, OP2):

        self.args = (OP1, OP2)
        _check_oneport_args(self.args)

        super(LSection, self).__init__(Series(OP1).chain(Shunt(OP2)))


class Ladder(TwoPortBModel):

    """(Unbalanced) ladder network with alternating Series and Shunt
    networks chained
    ::

           +---------+       +---------+
         --+   OP1   +---+---+ args[1] +---
           +---------+   |   +---------+
                       +-+-+
                       |   |
                       |   | args[0]
                       |   |
                       +-+-+
                         |
         ----------------+-----------------
    """

    def __init__(self, OP1, *args):

        self.args = (OP1, ) + args
        _check_oneport_args(self.args)

        TP = Series(OP1)

        for m, arg in enumerate(args):

            if m & 1:
                TP = TP.chain(Series(arg))
            else:
                TP = TP.chain(Shunt(arg))

        super(Ladder, self).__init__(TP)

    def simplify(self):

        if len(self.args) == 1:
            return Series(self.args[0])
        elif len(self.args) == 2:
            return LSection(*self.args)
        elif len(self.args) == 3:
            return TSection(*self.args)
        return self

        # A Ladder of voltage sources and current sources
        # collapses to a single Lsection comprised of the total
        # voltage and total current.


class GeneralTxLine(TwoPortBModel):

    """General transmission line

    Z0 is the (real) characteristic impedance (ohms)
    gamma is the propagation constant (1/m)
    l is the transmission line length (m)
    """

    def __init__(self, Z0, gamma, l):

        Z0 = LaplaceDomainExpression(Z0)
        gamma = LaplaceDomainExpression(gamma)
        l = ConstantDomainExpression(l)

        H = sym.exp(gamma * l)

        B11 = 0.5 * (H + 1 / H)
        B12 = 0.5 * (1 / H - H) * Z0
        B21 = 0.5 * (1 / H - H) / Z0
        B22 = 0.5 * (H + 1 / H)

        super(GeneralTxLine, self).__init__(BMatrix(((B11, B12), (B21, B22))))
        self.args = (Z0, gamma, l)


class LosslessTxLine(GeneralTxLine):

    """Losslees transmission line
        Z0 is the (real) characteristic impedance (ohms)
        c is the propagation speed (m/s)
        l is the transmission line length (m)
        """

    def __init__(self, Z0, c=1.5e8, l=1):

        gamma = s / c

        super(LosslessTxLine, self).__init__(Z0, gamma, l)


class TxLine(GeneralTxLine):

    """Transmission line

    R series resistance/metre
    L series inductance/metre
    G shunt conductance/metre
    C shunt capacitance/metre
    l is the transmission line length
    """

    def __init__(self, R, L, G, C, l=1):

        Z = R + s * L
        Y = G + s * C
        gamma = sym.sqrt(Z * Y)
        Z0 = sym.sqrt(Z / Y)

        super(TxLine, self).__init__(Z0, gamma, l)



"""This module provides the SubNetlist class.

Copyright 2020 Michael Hayes, UCECE

"""

from .mna import MNAMixin
from .netlistmixin import NetlistMixin
from .state import state
from .symbols import omega



class SubNetlist(NetlistMixin, MNAMixin):
    """This is a representation of a netlist for a particular
    transformation domain, such as ac, dc, transient, or noise."""

    def __new__(cls, netlist, kind):

        obj = netlist.select(kind=kind)
        # Need own context to avoid conflicts with Vn1 and Vn1(s), etc.
        obj.context = state.new_context()
        obj.kind = kind
        obj.__class__ = cls
        obj._analysis = obj.analyse()
        return obj

    def __init__(self, netlist, kind):
        """ kind can be 't', 'dc', 's', 'time', 'ivp', 'n*' or omega, 
        where 'n*' is a noise identifer and omega is an angular frequency."""
        
        if not isinstance(kind, str):
            return
        if kind[0] == 'n':
            return
        kinds = ('t', 'dc', 's', 'time', 'ivp', 'laplace')
        if kind not in kinds:
            raise ValueError('Expected one of %s for kind, got %s' %
                             (', '.join(kinds), kind))

    def get_I(self, name):
        """Current through component"""

        self._solve()
        return self._Idict[name].canonical()

    def get_i(self, name):
        """Time-domain current through component"""

        return self.get_I(name).time()

    def get_Vd(self, Np, Nm=None):
        """Voltage drop between nodes"""

        self._solve()
        return (self._Vdict[Np] - self._Vdict[Nm]).canonical()

    def get_vd(self, Np, Nm=None):
        """Time-domain voltage drop between nodes"""

        return self.get_Vd(Np, Nm).time()



"""
Lcapy is a Python library for linear circuit analysis.  It uses SymPy
for symbolic mathematics.

Lcapy can analyse circuits described with netlists using modified nodal
analysis.  See lcapy.netlist

Alternatively, Lcapy can analyse networks and circuits formed by
combining one, two, and three port networks.  See lcapy.oneport

For more detailed documentation see http://lcapy.elec.canterbury.ac.nz

Copyright 2014--2020 Michael Hayes, UCECE
"""
from __future__ import absolute_import, print_function
del absolute_import, print_function

name = "lcapy"

import pkg_resources

__version__ = pkg_resources.require('lcapy')[0].version
lcapy_version = __version__

import sys
if sys.version_info[0] == 2 and sys.version_info[1] < 6:
    raise ImportError("Python Version 2.6 or above is required for Lcapy.")
else:  # Python 3
    pass
    # Here we can also check for specific Python 3 versions, if needed

del sys

from .functions import *
from .symbols import *
from .circuit import *
from .oneport import *
from .twoport import *
from .schematic import *
from .expr import *
from .voltage import *
from .current import *
from .admittance import *
from .impedance import *
from .transfer import *
from .printing import *
from .sym import *
from .matrix import *
from .smatrix import *
from .tmatrix import *
from .vector import *
from .statespace import *
from .nodalanalysis import *
from .loopanalysis import *
from .laplace import *
from .nettransform import *
from .randomnetwork import *
from .simulator import *
from .fexpr import fexpr
from .sexpr import sexpr, zp2tf, tf, pr2tf
from .texpr import texpr
from .cexpr import cexpr
from .omegaexpr import omegaexpr
from .phasor import phasor
from .discretetime import *
from .dft import *

def show_version():
    """Show versions of Lcapy, SymPy, NumPy, MatplotLib, SciPy, and Python."""
    
    from sys import version as python_version
    from sympy import __version__ as sympy_version
    from numpy import __version__ as numpy_version
    from scipy import __version__ as scipy_version    
    from matplotlib import __version__ as matplotlib_version    

    print('Python: %s\nSymPy: %s\nNumPy: %s\nMatplotlib: %s\nSciPy: %s\nLcapy: %s' % 
          (python_version, sympy_version, numpy_version,
           matplotlib_version, scipy_version, lcapy_version))

# The following is to help sympify deal with j.   
# A better fix might be to define an Lcapy class for j and to
# use the __sympy_ method.
from sympy.core.sympify import converter
from sympy import Symbol
converter['j'] = j
converter[Symbol('j')] = j





"""This module provides the ExpressionClasses class, for dynamically
creating expression classes, and a factory to select the appropriate
class for a domain and quantity.

Copyright 2021 Michael Hayes, UCECE

"""

from .voltagemixin import VoltageMixin
from .currentmixin import CurrentMixin
from .admittancemixin import AdmittanceMixin
from .impedancemixin import ImpedanceMixin
from .transfermixin import TransferMixin
from .voltagesquaredmixin import VoltageSquaredMixin
from .currentsquaredmixin import CurrentSquaredMixin
from .admittancesquaredmixin import AdmittanceSquaredMixin
from .impedancesquaredmixin import ImpedanceSquaredMixin
from .powermixin import PowerMixin


quantityclasses = {'voltage': VoltageMixin,
                   'current': CurrentMixin,
                   'admittance': AdmittanceMixin,
                   'impedance': ImpedanceMixin,
                   'transfer': TransferMixin,
                   'voltagesquared': VoltageSquaredMixin,
                   'currentsquared': CurrentSquaredMixin,
                   'admittancesquared': AdmittanceSquaredMixin,
                   'impedancesquared': ImpedanceSquaredMixin,
                   'power': PowerMixin}


class ExpressionClassBuilder(dict):

    def __init__(self, domain, domainclass1, domainclass2=None, quantities=None):

        self.domain = domain
        self.domainclass1 = domainclass1
        self.domainclass2 = domainclass2
        self.quantities = quantities

    def make1(self, quantity, domainclass):

        if quantity == 'undefined':
            self[quantity] = domainclass
            return domainclass

        quantityclass = quantityclasses[quantity]
        quantityunits = quantityclass.units
            
        domainunits = domainclass.domain_units

        units = quantityunits            
        if quantity in ('voltage', 'current'):
            if (domainclass.is_laplace_domain or
                domainclass.is_fourier_domain or
                domainclass.is_angular_fourier_domain):
                units = '%s/%s' % (quantityunits, domainunits)
        elif quantity in ('voltagesquared', 'currentsquared'):
            if (domainclass.is_laplace_domain or
                domainclass.is_fourier_domain or
                domainclass.is_angular_fourier_domain):                
                units = '%s/%s^2' % (quantityunits, domainunits)
        elif quantity in ('impedance', 'admittance'):
            if domainclass.is_time_domain:
                units = '%s/%s' % (quantityunits, domainunits)
        elif quantity in ('admittancesquared', 'impedancesquared'):
            if domainclass.is_time_domain:
                units = '%s/%s^2' % (quantityunits, domainunits)

        # FIXME:  The units of squared quantities are incorrect under transformation
        # to another domain.  For example, v1(t) * v2(t) has units V^2,
        # V1(f) * V2(f) has units (V/Hz)^2, but (v1(t) * v2(t))(f) has units V^2/Hz,
        # and (V1(f) * V2(f))(t) has units V^2/Hz.
                        
        docstring = '%s-domain %s (units %s).' % (domainclass.domain_label,
                                                  quantity, units)

        suffix = 'Expression'
        name = domainclass.__name__.replace(suffix, quantityclass.quantity.capitalize())
            
        newclass = type(name, (quantityclass, domainclass),
                            {'__doc__': docstring,
                             'units': units})
        self[quantity] = newclass

        #print('Created %s %s' % (self.domain, quantity))
        return newclass
        
    def make(self, quantity):
        
        if self.quantities is None:
            return self.make1(quantity, self.domainclass1)

        if quantity in self.quantities:
            return self.make1(quantity, self.domainclass1)

        return self.make1(quantity, self.domainclass2)        

    def __getitem__(self, quantity):

        if quantity in self:
            return super(ExpressionClassBuilder, self).__getitem__(quantity)

        return self.make(quantity)
    

class ExpressionClasses(dict):                 
    
    def register(self, domain, domainclass1, domainclass2=None, quantities=None):

        self[domain] = ExpressionClassBuilder(domain, domainclass1, domainclass2, quantities)
        return self[domain]
                 
    def get_quantity(self, domain, quantity):

        return self[domain][quantity]


expressionclasses = ExpressionClasses()    



"""This module provides the Assumptions class.  It manages
assumptions for expressions needed for inverse Laplace transforms
and expression simplification.

Copyright 2014--2020 Michael Hayes, UCECE

"""

from copy import copy
from .acdc import is_dc, is_ac, is_causal


class Assumptions(dict):
    """Manage expression assumptions.  These include 'ac', 'dc', 'causal'
    and 'unknown' to specify expression behavior for t < 0, 'positive',
    'complex', etc., to help with expression simplification, 'nid' as a noise
    identifier, and `omega` for the phasor angular frequency.

    'ac', 'dc', 'causal' and 'unknown' refer to the time domain
    expression.  They can be inferred for a time domain expression but
    can be overridden.  The inference is performed lazily.  In
    particular, it is performed when converting from the time domain
    to the Laplace or Phasor domains.  The latter is required to
    handle the case of conversions from the Phasor to the Laplace
    domain.  Conversions from the Fourier domain to the Laplace domain
    go via the time domain and so the inference is performed while in
    the time domain.

    """

    def __init__(self, *args, **kwargs):

        super (Assumptions, self).__init__(*args)

        for assumption, value in kwargs.items():
            self.set(assumption, value)
    
    def set(self, assumption, value):

        if assumption in ('dc', 'ac', 'causal', 'unknown'):
            if value:
                self.pop('dc', None)
                self.pop('ac', None)
                self.pop('causal', None)
                self.pop('unknown', None)                
                self[assumption] = value
            else:
                self.pop(assumption, None)
        else:
            self[assumption] = value

    def get(self, assumption, default=False):
        
        return super(Assumptions, self).get(assumption, default)

    def merge(self, **assumptions):

        new = self.copy()
        # TODO: warn if have mutually exclusive assumptions.
        # Note, the last one overrides.
        for assumption, value in assumptions.items():
            new.set(assumption, value)
        return new

    def copy(self):

        return copy(self)
    
    def sympy_assumptions(self):
        """Return dict of the SymPy assumptions such as complex, positive, etc."""

        assumptions = {}
        for assumption, value in self.items():
            if assumption not in ('nid', 'ac', 'dc', 'causal', 'unknown'):
                assumptions[assumption] = value
        return assumptions

    def infer_from_expr(self, expr):

        if expr.is_transform_domain:
            self.set('unknown', True)
            return
        
        var = expr.var
        if is_dc(expr, var):
            self.set('dc', True)
            return

        if is_ac(expr, var):
            self.set('ac', True)            
            return

        if is_causal(expr, var):
            self.set('causal', True)            
            return

        self.set('unknown', True)                    

    @property
    def has_unspecified(self):

        if 'ac' in self:
            return False
        if 'dc' in self:
            return False
        if 'causal' in self:
            return False
        if 'unknown' in self:
            return False        
        return True

    def merge_and_infer(self, expr, **assumptions):
        """Override assumptions with specified assumptions.
        If none specified, infer them."""

        assumptions = self.merge(**assumptions)        
        if assumptions.has_unspecified:
            assumptions.infer_from_expr(expr)
        return assumptions

    @property
    def is_ac(self):

        return self.get('ac')

    @property
    def is_dc(self):

        return self.get('dc')

    @property
    def is_causal(self):

        return self.get('causal')

    @property
    def is_unknown(self):

        return self.get('unknown')    

    def convolve(self, x):

        # The assumptions refer to the time-domain signal or
        # impulse-response.  Thus we want to propagate the assumptions
        # for the voltage or current signal.  The ac, dc, and causal
        # assumptions are only required for s-domain expressions.  For
        # other signals they can be determined from the time-domain
        # response.
        
        assumptions = self.copy()
        assumptions.set('unknown', True)
        
        if self.is_unknown or x.is_unknown:
            assumptions.set('unknown', True)
        elif self.is_ac or x.is_ac:
            assumptions.set('ac', True)            
        elif self.is_dc or x.is_dc:
            assumptions.set('dc', True)                        
        elif self.is_causal or x.is_causal:
            assumptions.set('causal', True)                        
        return assumptions

    def add(self, x):

        assumptions = self.copy()
        assumptions.set('unknown', True)
        
        if self.is_causal and x.is_causal:
            assumptions.set('causal', True)                                    
        elif self.is_dc and x.is_dc:
            assumptions.set('dc', True)
        elif self.is_ac and x.is_ac:
            assumptions.set('ac', True)            
        return assumptions    
    



from lcapy import *
from lcapy.discretetime import *
import unittest
import sympy as sym


class LcapyTester(unittest.TestCase):

    """Unit tests for lcapy

    """

    def test_transform(self):

        a = ui(n)

        self.assertEqual(a(z), 1, "ui(n)")

    def test_decompose_AB(self):

        H = (z -2) / (z - 3)
        A, B = H.decompose_AB()

        self.assertEqual(A, 1 - 3 / z, "A")
        self.assertEqual(B, 1 - 2 / z, "A")                

    def test_sequence(self):

        x = seq((1, 2, 3))
        e = delta(n) + 2 * delta(n - 1) + + 3 * delta(n - 2)

        y = seq((1, 2, 3, 0, 0))

        h = seq((1, 2))                
        
        self.assertEqual(x.as_impulses(), e, "as_impulses")
        self.assertEqual(x.zeropad(2), y, "zeropad")
        self.assertEqual(x.prune(), x, "prune")
        self.assertEqual(y.extent(), 3, "extent")
        self.assertEqual(x.convolve(h), seq((1, 4, 7, 6)), "convolve")
        self.assertEqual(h.ZT(), (z + 2) / z, "ZT")
        self.assertEqual(h.latex(), r'\left\{\underline{1}, 2\right\}', "latex")                
        self.assertEqual(seq('1, 2, 3'), x, "seq")
        self.assertEqual(seq('{1, _2, 3}').latex(), r'\left\{1, \underline{2}, 3\right\}', "latex")
        self.assertEqual(e.seq(), x, "seq")        
        
    def test_zexpr(self):

         H = z / (z - 1)

         DE = H.difference_equation()
         rhs = expr('x(n)')
         lhs = expr('y(n) - y(n-1)')

         self.assertEqual(DE.lhs, lhs, "DE lhs")
         self.assertEqual(DE.rhs, rhs, "DE rhs")                  

    def test_nexpr(self):

        x = seq('{1, _2, 3, 4}').as_impulses()

        y1 = seq('{0, 1, 2, 3, 4}').as_impulses()        
        y2 = x.delay(2)

        self.assertEqual(y1, y2, "delay(2)")        
        self.assertEqual(x.first_index(), -1, "first_index")
        self.assertEqual(x.last_index(), 2, "last_index")        



"""
This module provides a class to represent circuits as graphs.
This is primarily for loop analysis but is also used for nodal analysis.

Copyright 2019--2020 Michael Hayes, UCECE

"""

from matplotlib.pyplot import subplots, savefig
import networkx as nx


# V1 1 0 {u(t)}; down
# R1 1 2; right=2
# L1 2 3; down=2
# W1 0 3; right
# W 1 5; up
# W 2 6; up
# C1 5 6; right=2

class CircuitGraph(nx.Graph):

    def __init__(self, cct):

        super(CircuitGraph, self).__init__()
        self.cct = cct
        self.dummy = 0
        # Dummy nodes are used to avoid parallel edges.
        self.dummy_nodes = {}

        self.add_nodes_from(cct.node_list)

        node_map = cct.node_map

        for name in cct.branch_list:
            elt = cct.elements[name]
            if len(elt.nodenames) < 2:
                continue

            nodename1 = node_map[elt.nodenames[0]]
            nodename2 = node_map[elt.nodenames[1]]
            
            if self.has_edge(nodename1, nodename2):
                # Add dummy node in graph to avoid parallel edges.
                dummynode = '*%d' % self.dummy
                dummycpt = 'W%d' % self.dummy                
                self.add_edge(nodename1, dummynode, name=name)
                self.add_edge(dummynode, nodename2, name=dummycpt)
                self.dummy_nodes[dummynode] = nodename2
                self.dummy += 1
            else:
                self.add_edge(nodename1, nodename2, name=name)

        self.node_map = node_map

    def connected(self, node):
        """List of components connected to specified node."""

        for node1, edges in self.node_edges(node).items():
            if node1.startswith('*'):
                for elt in self.connected(node1):
                    yield elt
                continue
            
            for key, edge in edges.items():
                if not edge.startswith('W'):
                    elt = self.cct.elements[edge]
                    yield elt
            
    def all_loops(self):

        # This adds forward and backward edges.
        DG = nx.DiGraph(self)
        cycles = list(nx.simple_cycles(DG))

        loops = []
        for cycle in cycles:
            if len(cycle) <= 2:
                continue
            cycle = sorted(cycle)
            if cycle not in loops:
                loops.append(cycle)        
        return loops

    def chordless_loops(self):

        loops = self.all_loops()
        sets = [set(loop) for loop in loops]

        DG = nx.DiGraph(self)
        
        rejects = []
        for i in range(len(sets)):

            # Reject loops with chords.
            loop = loops[i]
            if len(loop) == 2:
                continue
            
            for j in range(i + 1, len(sets)):
                if sets[i].issubset(sets[j]):
                    rejects.append(j)
                elif sets[j].issubset(sets[i]):
                    rejects.append(i)

        cloops = []
        for i, loop in enumerate(loops):
            if i not in rejects:
                cloops.append(loop)

        return cloops

    def loops(self):
        if hasattr(self, '_loops'):
            return self._loops
        self._loops = self.chordless_loops()
        return self._loops
    
    def draw(self, filename=None):
        """Use matplotlib to draw circuit graph."""

        fig, ax = subplots(1)

        G = self
        pos = nx.spring_layout(G)
        
        labels = dict(zip(G.nodes(), G.nodes()))
        nx.draw_networkx(G, pos, ax, labels=labels)
        
        edge_labels = dict([((u, v), d['name'])
                            for u, v, d in G.edges(data=True)])
        nx.draw_networkx_edge_labels(G, pos, edge_labels=edge_labels)

        if filename is not None:
            savefig(filename, bbox_inches='tight')

    @property
    def is_planar(self):
        """Return True for a planar network."""
        
        return nx.check_planarity(self)[0]

    def node_edges(self, node):

        return self[node]

    def component(self, node1, node2):

        name = self.get_edge_data(node1, node2)['name']
        if name.startswith('W'):
            return None
        return self.cct.elements[name]

    def loops_for_cpt(self, elt):
        """Return list of tuples; one for each loop.  The first element of the
        tuple is the loop the cpt belongs to or an empty list; the
        second element indicates the cpt direction compared to the
        loop direction."""
        
        loops = self.loops()
        cloops = []

        # Map node names to equipotential node names.
        nodenames = [self.cct.node_map[nodename] for nodename in elt.nodenames]
        
        for n, loop in enumerate(loops):

            loop1 = loop.copy()
            loop1.append(loop1[0])

            def find(loop1, nodename1, nodename2):
                for m in range(len(loop1) - 1):
                    if (nodename1 == loop1[m] and
                        nodename2 == loop1[m + 1]):
                        return True

            if find(loop1, nodenames[0], nodenames[1]):
                cloops.append((loop, False))
            elif find(loop1, nodenames[1], nodenames[0]):
                cloops.append((loop, True))
            else:
                cloops.append(([], None))

        return cloops
    
    @property
    def components(self):
        """Return list of component names."""

        return [d['name'] for n1, n2, d in self.edges(data=True)]

    def in_series(self, cpt_name):
        """Return set of component names in series with cpt including itself."""        

        cct = self.cct
        elt = cct.elements[cpt_name]
        nodenames = [cct.node_map[nodename] for nodename in elt.nodenames]

        series = []
        series.append(cpt_name)        

        def follow(node):
            neighbours = self[node]
            if len(neighbours) > 2:
                return
            for n, e in neighbours.items():
                if not e['name'] in series:
                    series.append(e['name'])
                    follow(n)

        follow(nodenames[0])
        follow(nodenames[1])        

        # If only have two components in parallel, they will be
        # detected as a series connection.  However, if there is a
        # dummy wire, the components are in parallel.
        for name in series:
            if name.startswith('W'):
                return set((cpt_name, ))

        return set(series)

    def in_parallel(self, cpt_name):
        """Return set of component names in parallel with cpt including itself."""        

        cct = self.cct
        elt = cct.elements[cpt_name]
        nodenames = [cct.node_map[nodename] for nodename in elt.nodenames]

        n1, n2 = nodenames[0:2]

        # This is trivial for a multigraph but a mutigraph adds additional problems
        # since component() will fail if have multiple edges between the same nodes.
        # edges = self.get_edge_data(n1, n2)
        # parallel = [d['name'] for k, d in edges.items()]

        parallel = [cpt_name]

        neighbours1 = self[n1]
        neighbours2 = self[n2]

        # The first created parallel component has no dummy nodes.
        try:
            name = self.get_edge_data(n1, n2)['name']
            parallel.append(name)
        except:
            pass
            
        # If find a dummy node name then there is a parallel component.
        
        for n, e in neighbours1.items():
            if n.startswith('*'):
                for n3, e3 in self[n].items():
                    if n3 == n2 and not e['name'].startswith('W'):
                        parallel.append(e['name'])
        for n, e in neighbours2.items():
            if n.startswith('*'):
                for n3, e3 in self[n].items():
                    if n3 == n1 and not e['name'].startswith('W'):
                        parallel.append(e['name'])

        if n1.startswith('*'):
            for n3, e3 in self[n1].items():
                if n3 == n2 and not e['name'].startswith('W'):
                    parallel.append(e['name'])
        if n2.startswith('*'):
            for n3, e3 in self[n2].items():
                if n3 == n1 and not e['name'].startswith('W'):
                    parallel.append(e['name'])                                            
        
        return set(parallel)




"""This module provides the ConstantDomainExpression class to represent constant expressions.

Copyright 2014--2020 Michael Hayes, UCECE

"""

from .expr import Expr
from .domains import ConstantDomain
from .sym import symbols_find
from .voltagemixin import VoltageMixin
from .currentmixin import CurrentMixin
from .admittancemixin import AdmittanceMixin
from .impedancemixin import ImpedanceMixin
from .transfermixin import TransferMixin

class ConstantDomainExpression(ConstantDomain, Expr):
    """Constant real expression or symbol.

    If symbols in the expression are known to be negative, use
    ConstantDomainExpression(expr, positive=False)

    """

    def __init__(self, val, **assumptions):

        symbols = symbols_find(val)
        for symbol in ('s', 'omega', 't', 'f'):
            if symbol in symbols:
                raise ValueError(
                    'constant expression %s cannot depend on %s' % (val, symbol))

        assumptions['dc'] = True        
        super(ConstantDomainExpression, self).__init__(val, **assumptions)

    def as_expr(self):
        return ConstantDomainExpression(self)

    def rms(self):
        """Return root mean square."""
        return self

    def fourier(self):
        """Convert to Fourier domain representation."""

        return self.time().fourier()

    def angular_fourier(self):
        """Convert to angular Fourier domain representation."""

        return self.time().angular_fourier()        

    def canonical(self, factor_const=True):
        # Minor optimisation
        return self

    def phasor(self, **assumptions):

        return self.laplace(**assumptions).phasor(**assumptions)

    def time(self, **assumptions):
        return self.change(self, domain='time', **assumptions)

    def laplace(self):
        """Convert to Laplace domain representation."""

        return self.time().laplace()    
    

class ConstantTimeDomainExpression(ConstantDomainExpression):
    pass
    

class ConstantFrequencyDomainExpression(ConstantDomainExpression):

    def laplace(self, **assumptions):
        return self.change(self, domain='laplace', **assumptions)

    def time(self, **assumptions):
        """Convert to time domain."""
        
        return self.laplace(**assumptions).time(**assumptions)        

    
def cexpr(arg, **assumptions):
    """Create Lcapy constant expression from `arg`.

    By default, `arg` is assumed to be positive.  If symbols in the
    `arg` are known to be negative, use `cexpr(arg, positive=False)`.

    """

    return ConstantDomainExpression(arg)


from .expressionclasses import expressionclasses

expressionclasses.register('constant', ConstantTimeDomainExpression, ConstantFrequencyDomainExpression,
                           ('voltage', 'current', 'voltagesquared', 'currentsquared'))



"""This module provides impedance support.

Copyright 2019--2020 Michael Hayes, UCECE

"""
from __future__ import division
from .expr import expr
from .sexpr import LaplaceDomainExpression
from .units import u as uu


def impedance(arg, **assumptions):
    """Generic impedance factory function.

    Y(omega) = G(omega) + j * B(omega)

    where G is the conductance and B is the susceptance.

    Admittance is the reciprocal of impedance,

    Z(omega) = 1 / Y(omega)

    """

    expr1 = expr(arg, **assumptions)    

    try:
        expr1 = expr1.as_impedance()
    except:    
        raise ValueError('Cannot represent %s(%s) as impedance' % (expr1.__class__.__name__, expr1))
        
    return expr1.apply_unit(uu.ohms)    
    



from lcapy import *
from lcapy.omegaexpr import AngularFourierDomainExpression
from lcapy.texpr import TimeDomainExpression
import unittest


class LcapyTester(unittest.TestCase):
    """Unit tests for lcapy"""

    def assertEqual2(self, ans1, ans2, comment):

        try:
            self.assertEqual(ans1, ans2, comment)
        except AssertionError as e:
            pprint(ans1)
            pprint(ans2)
            raise AssertionError(e)

    def test_tdomain_mul(self):

        v = voltage(5 * t)
        self.assertEqual(v * 3, 15 * t, "v(t) * const")
        self.assertEqual(3 * v, 15 * t, "const * v(t)")

    def test_sdomain_mul(self):
        
        V = voltage(5 / s)
        I = current(sexpr(5))        
        Y = admittance(s)
        Z = impedance(1 / s)        
        H = transfer(sexpr(10))
        U = sexpr(2)
        C = cexpr(3)
        
        self.assertEqual(V * 3, 15 / s, "V(s) * const")
        self.assertEqual(V * C, 15 / s, "V(s) * const")        
        self.assertEqual(3 * V, 15 / s, "const * V(s)")
        self.assertEqual(C * V, 15 / s, "const * V(s)")        
        self.assertEqual(Y * V, I, "Y(s) * V(s)")
        self.assertEqual(V * Y, I, "V(s) * Y(s)")
        self.assertEqual(I * Z, V, "I(s) * Z(s)")
        self.assertEqual(Z * I, V, "Z(s) * I(s)")                
        self.assertEqual(V * H, voltage(50 / s), "V(s) * H(s)")
        self.assertEqual(H * V, voltage(50 / s), "H(s) * V(s)")
        self.assertEqual(I * H, current(sexpr(50)), "I(s) * H(s)")
        self.assertEqual(H * I, current(sexpr(50)), "H(s) * I(s)")
        self.assertEqual(U * U, sexpr(4), "U(s) * U(s)")        

    def test_fdomain_mul(self):
        
        V = voltage(5 / f)
        I = current(fexpr(5))        
        Y = admittance(f)
        Z = impedance(1 / f)                
        H = transfer(fexpr(10))
        U = fexpr(2)
        C = cexpr(3)
        
        self.assertEqual(V * 3, 15 / f, "V(f) * const")
        self.assertEqual(V * C, 15 / f, "V(f) * const")
        self.assertEqual(3 * V, 15 / f, "const * V(f)")
        self.assertEqual(C * V, 15 / f, "const * V(f)")        
        self.assertEqual(Y * V, I, "Y(f) * V(f)")
        self.assertEqual(V * Y, I, "V(f) * Y(f)")
        self.assertEqual(I * Z, V, "I(f) * Z(f)")
        self.assertEqual(Z * I, V, "Z(f) * I(f)")                        
        self.assertEqual(V * H, voltage(50 / f), "V(f) * H(f)")
        self.assertEqual(H * V, voltage(50 / f), "H(f) * V(f)")
        self.assertEqual(I * H, current(fexpr(50)), "I(f) * H(f)")
        self.assertEqual(H * I, current(fexpr(50)), "H(f) * I(f)")
        self.assertEqual(U * U, fexpr(4), "U(f) * U(f)")

    def test_mixed_mul(self):
        
        self.assertEqual(type(omega * t), TimeDomainExpression, "omega * t")
        self.assertEqual(type(t * omega), TimeDomainExpression, "omega * t")

    def test_sdomain_div(self):
        
        V = voltage(5 / s)
        I = current(sexpr(5))        
        Y = admittance(s)
        Z = impedance(1 / s)        
        H = transfer(sexpr(10))
        U = sexpr(2)        
        
        self.assertEqual(V / 5, 1 / s, "V(s) / 5")
        self.assertEqual(V / Z, I, "V(s) / Z(s)")
        self.assertEqual(I / Y, V, "I(s) / Y(s)")
        self.assertEqual(I / H, current(sexpr(1 / 2)), "I(s) / H(s)")
        self.assertEqual(V / H, voltage(sexpr(1 / 2 / s)), "V(s) / H(s)")
        self.assertEqual(U / U, sexpr(1), "U(s) / U(s)")
        self.assertEqual(Z / Z, sexpr(1), "Z(s) / Z(s)")
        self.assertEqual(Y / Y, sexpr(1), "Y(s) / Y(s)")                        

    def test_sdomain_rdiv(self):
        
        Y = admittance(s)
        Z = impedance(1 / s)        
        H = transfer(sexpr(10))
        C = cexpr(1)        
        
        self.assertEqual(1 / Z, Y, "1 / Z(s)")
        self.assertEqual(1 / Y, Z, "1 / Y(s)")
        self.assertEqual(C / Z, Y, "C / Z(s)")
        self.assertEqual(C / Y, Z, "C / Y(s)")         

    def test_sdomain_add(self):
        
        V1 = voltage(s)
        V2 = voltage(s + 4)
        C = cexpr(3)
        Z = impedance(3 * s)
        CZ = impedance(3)        
        
        self.assertEqual(V1 + V2, voltage(2 * s + 4), "V1(s) + V2(s)")
        self.assertEqual(C + V1, voltage(3 + s), "C + V1(s)")
        self.assertEqual(V1 + C, voltage(3 + s), "V1(s) + C")
        self.assertEqual(V1 + 3, voltage(3 + s), "V1(s) + 3")
        self.assertEqual(3 + V1, voltage(3 + s), "3 + V1(s)")
        self.assertEqual(C + Z, impedance(3 * s + 3), "C + Z(s)")
        self.assertEqual(Z + C, impedance(3 * s + 3), "Z(s) + C")
        self.assertEqual(CZ + Z, impedance(3 * s + 3), "CZ + Z(s)")
        self.assertEqual(Z + CZ, impedance(3 * s + 3), "Z(s) + CZ")        
        
    def test_fdomain_add(self):
        
        V1 = voltage(f)
        V2 = voltage(f + 4)
        C = cexpr(3)
        
        self.assertEqual(V1 + V2, voltage(2 * f + 4), "V1(f) + V2(f)")
        self.assertEqual(C + V1, voltage(3 + f), "C + V1(f)")
        self.assertEqual(V1 + C, voltage(3 + f), "V1(f) + C")
        self.assertEqual(V1 + 3, voltage(3 + f), "V1(f) + 3")
        self.assertEqual(3 + V1, voltage(3 + f), "3 + V1(f)")                                        
        
    def test_rmul(self):

        self.assertEqual2((j * omega).__class__,  AngularFourierDomainExpression, "j * omega fail.")

    def test_comparison(self):

        a = expr('3')
        b = expr('4')
        self.assertEqual(a < b, True, "a < b")
        self.assertEqual(a > b, False, "a > b")
        self.assertEqual(a <= b, True, "a <= b")
        self.assertEqual(a >= b, False, "a >= b")        
        self.assertEqual(a == b, False, "a == b")
        self.assertEqual(a != b, True, "a != b")        

        a = voltage(3)
        b = current(3)
        self.assertEqual(a != b, True, "a != b")

        a = voltage(3 * s)
        b = voltage(3)
        self.assertEqual(a != b, True, "a != b")

        self.assertEqual(jw != s, True, "jw != s")                

    def test_rdiv(self):

        self.assertEqual(3 / expr('3'), 1, "3 / expr('3')")

    def test_div(self):

        self.assertEqual(expr('3') / 3, 1, "expr('3') / 3")        
        
    def test_rsub(self):

        self.assertEqual(3 - expr('3'), 0, "3 - expr('3')")

    def test_sub(self):

        self.assertEqual(expr('3') - 3, 0, "expr('3') - 3")

    def test_time_op_const(self):

        self.assertEqual(voltage(texpr(10)) / impedance(2), current(5), 'V / R')
        self.assertEqual(current(texpr(5)) / admittance(1 / 2), voltage(10), 'V / G')
        self.assertEqual(voltage(texpr(10)) * admittance(1 / 2), current(5), 'V * Y')
        self.assertEqual(current(texpr(5)) * impedance(2), voltage(10), 'I * Z')

        # Invoke conversion of LaplaceDomain to ConstantDomain
        self.assertEqual(current(texpr(5)) * impedance(0 * s + 2), voltage(10), 'I * Z')
        self.assertEqual(voltage(texpr(10)) / impedance(0 * s + 2), current(5), 'V / Z')                
        

        



from lcapy import *
from lcapy.phasor import PhasorTimeDomainExpression
from lcapy.sexpr import LaplaceDomainVoltage
import unittest
import sympy as sym
from lcapy.sym import omega0sym

class LcapyTester(unittest.TestCase):

    """Unit tests for lcapy

    """

    def assertEqual2(self, ans1, ans2, comment):

        try:
            self.assertEqual(ans1, ans2, comment)
        except AssertionError as e:
            ans1.pprint()
            ans2.pprint()
            raise AssertionError(e)

    def test_R(self):
        """Lcapy: check R

        """
        a = R(10)
        self.assertEqual(a.Z, 10, "Z incorrect.")
        self.assertEqual(a.Y * 10, 1, "Y incorrect.")        
        self.assertEqual2(a.Voc.s, 0, "Voc incorrect.")
        self.assertEqual2(a.Isc.s, 0, "Isc incorrect.")
        self.assertEqual2(a.voc, 0, "voc incorrect.")
        self.assertEqual2(a.isc, 0, "isc incorrect.")

    def test_L(self):
        """Lcapy: check L

        """
        a = L(10, 5)
        self.assertEqual2(a.Z(s), 10 * s, "Zs incorrect.")
        self.assertEqual2(a.Y(s), 1 / (10 * s), "Ys incorrect.")
        self.assertEqual2(a.Z(jw), 10 * j * omega, "Z incorrect.")
        self.assertEqual2(a.Y(jw), 1 / (10 * j * omega), "Y incorrect.")                
        self.assertEqual2(a.Voc.s, -10 * 5, "Voc incorrect.")
        self.assertEqual2(a.Isc.s, -5 / s, "Isc incorrect.")        

    def test_C(self):
        """Lcapy: check C

        """
        a = C(10, 5)

        self.assertEqual2(a.Z(s), 1 / (10 * s), "Zs incorrect.")
        self.assertEqual2(a.Y(s), 10 * s, "Ys incorrect.")
        self.assertEqual2(a.Z(jw), 1 / (10 * j * omega), "Z incorrect.")
        self.assertEqual2(a.Y(jw), 10 * j * omega, "Y incorrect.")                
        self.assertEqual2(a.Voc.s, 5 / s, "Voc incorrect.")
        self.assertEqual(a.Isc.s, 50, "Isc incorrect.")        

    def test_R_series_R(self):
        """Lcapy: check R + R

        """
        a = R(10) + R(5)
        b = a.simplify()
        self.assertEqual2(b.R, 15, "R incorrect.")
        self.assertEqual2(b.Z, 15, "Z incorrect.")
        self.assertEqual(b.Voc.s, 0, "Voc incorrect.")
        self.assertEqual(b.Isc.s, 0, "Isc incorrect.")                
        self.assertEqual2(type(b), R, "type incorrect.")

    def test_L_series_L(self):
        """Lcapy: check L + L

        """
        a = L(10) + L(5)
        b = a.simplify()
        self.assertEqual2(b.L, 15, "L incorrect.")
        self.assertEqual2(b.Z(s), 15 * s, "Zs incorrect.")
        self.assertEqual2(b.Z(jw), 15 * j * omega, "Z incorrect.")        
        self.assertEqual2(type(b), L, "type incorrect.")

    def test_C_series_C(self):
        """Lcapy: check C + C

        """
        a = C(10) + C(15)
        b = a.simplify()
        self.assertEqual2(b.C, 6, "C incorrect.")
        self.assertEqual2(b.Z(s), 1 / (6 * s), "Zs incorrect.")
        self.assertEqual2(b.Z(jw), 1 / (6 * j * omega), "Z incorrect.")        
        self.assertEqual2(type(b), C, "type incorrect.")

    def test_Vdc_series_Vdc(self):
        """Lcapy: check Vdc + Vdc

        """
        a = Vdc(10) + Vdc(5)
        b = a.simplify()
        self.assertEqual2(b.voc, Vdc(15).voc, "voc incorrect.")
        self.assertEqual(b.Voc.dc, 15, "Voc incorrect.")
        self.assertEqual2(type(b), Vdc, "type incorrect.")
        self.assertEqual2(b.voc, 15, "voc incorrect.")

    def test_Vdc_series_R(self):
        """Lcapy: check Vdc + R

        """
        a = Vdc(10) + R(5)
        self.assertEqual2(a.voc, Vdc(10).voc, "voc incorrect.")
        self.assertEqual(a.Voc.dc, 10, "Voc incorrect.")
        self.assertEqual(a.V.dc, 10, "Voc incorrect.")
        self.assertEqual(a.I, 0, "I incorrect.")
        self.assertEqual(a.v(0), 10, "Voc incorrect.")                
        self.assertEqual(a.i(0), 0, "I incorrect.")                

        b = a.norton()
        self.assertEqual2(b.voc, Vdc(10).voc, "voc incorrect.")
        self.assertEqual(b.Voc.dc, 10, "Voc incorrect.")
        self.assertEqual(b.V.dc, 10, "Voc incorrect.")
        self.assertEqual(b.I, 0, "I incorrect.")
        self.assertEqual(b.v(0), 10, "Voc incorrect.")                
        self.assertEqual(b.i(0), 0, "I incorrect.")

    def test_Vac_series_R(self):
        """Lcapy: check Vac + R

        """
        a = Vac(10) + R(5)
        self.assertEqual2(a.voc, Vac(10).voc, "voc incorrect.")
        self.assertEqual(a.Voc.ac[omega0sym], PhasorTimeDomainExpression(10), "Voc incorrect.")
        self.assertEqual(a.V.ac[omega0sym], PhasorTimeDomainExpression(10), "Voc incorrect.")
        self.assertEqual(a.I, 0, "I incorrect.")
        self.assertEqual(a.v(0), 10, "Voc incorrect.")                
        self.assertEqual(a.i(0), 0, "I incorrect.")                

        b = a.norton()
        self.assertEqual2(b.voc, Vac(10).voc, "voc incorrect.")
        self.assertEqual(b.Voc.ac[omega0sym], PhasorTimeDomainExpression(10), "Voc incorrect.")
        self.assertEqual(b.V.ac[omega0sym], PhasorTimeDomainExpression(10), "Voc incorrect.")
        self.assertEqual(b.I, 0, "I incorrect.")
        self.assertEqual(b.v(0), 10, "Voc incorrect.")                
        self.assertEqual(b.i(0), 0, "I incorrect.")        
        
    def test_R_series_L(self):
        """Lcapy: check R + L

        """
        a = R(10) + L(5)
        self.assertEqual2(a.Z(s), 5 * s + 10, "Zs incorrect.")

    def test_R_parallel_R(self):
        """Lcapy: check R | R

        """
        a = R(10) | R(15)
        b = a.simplify()
        self.assertEqual2(b.R, 6, "R incorrect.")
        self.assertEqual2(b.Z, 6, "Z incorrect.")
        self.assertEqual2(type(b), R, "type incorrect.")

    def test_L_parallel_L(self):
        """Lcapy: check L | L

        """
        a = L(10) | L(15)
        b = a.simplify()
        self.assertEqual2(b.L, 6, "L incorrect.")
        self.assertEqual2(b.Z(s), 6 * s, "Zs incorrect.")
        self.assertEqual2(type(b), L, "type incorrect.")

    def test_C_parallel_C(self):
        """Lcapy: check C | C

        """
        a = C(10) | C(15)
        b = a.simplify()
        self.assertEqual2(b.C, 25, "C incorrect.")
        self.assertEqual2(b.Z(s), 1 / (25 * s), "Zs incorrect.")
        self.assertEqual2(type(b), C, "type incorrect.")

    def test_I_parallel_I(self):
        """Lcapy: check I | I

        """
        a = Idc(10) | Idc(5)
        b = a.simplify()
        self.assertEqual2(b.isc, Idc(15).isc, "isc incorrect.")
        self.assertEqual2(b.Isc.dc, 15, "Isc incorrect.")
        self.assertEqual2(type(b), Idc, "type incorrect.")
        self.assertEqual2(b.isc, 15, "isc incorrect.")

    def test_load(self):
        """Lcapy: check load

        """

        a = Shunt(R(10) + Vdc(5))
        b = a.load(R(30))

        self.assertEqual2(type(b), Ser, "type incorrect.")
        self.assertEqual2(b.Z, 7.5, "Shunt loaded R incorrect Z.")

    def test_open_circuit(self):
        """Lcapy: check open_circuit

        """

        a = Shunt(R(10) + Vdc(5))
        b = a.open_circuit()

        self.assertEqual2(b.Z, 10, "incorrect Z.")
        self.assertEqual2(b.Voc.laplace(), LaplaceDomainVoltage(5) / s, "incorrect V.")

    def test_short_circuit(self):
        """Lcapy: check short_circuit

        """

        a = Series(R(10) + Vdc(5))
        b = a.short_circuit()

        self.assertEqual2(b.Z, 10, "incorrect Z.")
        self.assertEqual2(b.Voc.laplace(), LaplaceDomainVoltage(5) / s, "incorrect V.")

    def test_v(self):
        """Lcapy: check inverse Laplace for voltage sources"""

        omega0 = symbol('omega_0', real=True)

        a = Vdc(10)
        self.assertEqual(a.voc, 10, "DC incorrect.")        

        a = Vstep(10)
        self.assertEqual2(a.voc, 10 * Heaviside(t), "Step incorrect.")

        a = Vac(10)
        self.assertEqual2(a.voc, 10 * cos(omega0 * t), "AC incorrect.")


    def test_i(self):
        """Lcapy: check inverse Laplace for current sources"""

        omega0 = symbol('omega_0', real=True)

        a = Idc(10)
        self.assertEqual(a.isc, 10, "DC incorrect.")        

        a = Istep(10)
        self.assertEqual2(a.isc, 10 * Heaviside(t), "Step incorrect.")

        a = Iac(10)
        self.assertEqual2(a.isc, 10 * cos(omega0 * t), "AC incorrect.")


    def test_CPE(self):
        """Lcapy: check CPE

        """
        a = CPE(10, 1)

        self.assertEqual2(a.Z(s), 1 / (10 * s), "Zs incorrect.")
        self.assertEqual2(a.Y(s), 10 * s, "Ys incorrect.")
        self.assertEqual2(a.Z(jw), 1 / (10 * j * omega), "Zw incorrect.")
        self.assertEqual2(a.Y(jw), 10 * j * omega, "Yw incorrect.")

        a = CPE(10, 0)

        self.assertEqual2(a.Z(s), 1.0 / 10, "Zs incorrect.")
        self.assertEqual2(a.Y(s), 10, "Ys incorrect.")
        self.assertEqual2(a.Z(jw), 1.0 / 10, "Zw incorrect.")
        self.assertEqual2(a.Y(jw), 10, "Yw incorrect.")

        a = CPE(10, 0.5)

        self.assertEqual2(a.Y(s), 10 * sqrt(s), "Ys incorrect.")

    def simplify1(self):

        a = I(0) + Y(1)
        b = a.simplify()

        self.assertEqual2(b._Z, None, "Z incorrect.")
        self.assertEqual2(b._Y, 1, "Y incorrect.")                

    def test_cpt(self):

        self.assertEqual(type(voltage(2).cpt()), Vdc , "-> Vdc")
        #self.assertEqual(type(voltage(cos(t)).cpt()), Vac, "-> Vac")

        self.assertEqual(type(current(2).cpt()), Idc, "-> Idc")
        #self.assertEqual(type(current(cos(t)).cpt()), Iac, "-> Iac")
        
        self.assertEqual(type(impedance(2).cpt()), R , "-> R")
        self.assertEqual(type(impedance(2 * s).cpt()), L, "-> L")
        self.assertEqual(type(impedance(2 / s).cpt()), C, "-> C")

        self.assertEqual(type(admittance(2).cpt()), G, "-> G")
        self.assertEqual(type(admittance(2 * s).cpt()), C, "-> C")
        self.assertEqual(type(admittance(2 / s).cpt()), L, "-> L")        
        
        



"""
This module performs loop analysis.  It is primarily for showing
the equations rather than evaluating them.

Copyright 2019--2020 Michael Hayes, UCECE

"""

# TODO, handle current sources and dependent sources.

from .circuitgraph import CircuitGraph
from .expr import expr
from .utils import equation
from .systemequations import SystemEquations
import sympy as sym


class LoopAnalysis(object):
    """
    This is an experimental class for loop analysis.  The API
    is likely to change.

    >>> from lcapy import Circuit, LoopAnalysis
    >>> cct = Circuit('''
    ... V1 1 0 {u(t)}; down
    ... R1 1 2; right=2
    ... L1 2 3; down=2
    ... W1 0 3; right
    ... W 1 5; up
    ... W 2 6; up
    ... C1 5 6; right=2
    ...''')    

    To perform nodal analysis in the time domain:

    >>> la = LoopAnalysis(cct)
    
    To display the equations found by applying KVL around each mesh:
    
    >>> la.mesh_equations().pprint()

    """
    
    def __init__(self, cct):

        self.cct = cct
        self.G = CircuitGraph(cct)

        self.kind = self.cct.kind
        if self.kind == 'super':
            self.kind = 'time'

        self._equations = self._make_equations()

        self._unknowns = self.mesh_currents()
        self._y = matrix(self._unknowns)

    def loops(self):

        return self.G.loops()

    @property
    def num_loops(self):

        return len(self.G.loops)

    def mesh_currents(self):

        if not self.G.is_planar:
            raise ValueError('Circuit topology is not planar')

        loops = self.loops()
        Nloops = len(loops)
        
        mesh_currents = ExprList([Iname('I_%d' % (m + 1), self.kind) for m in range(Nloops)])
        return mesh_currents

    def _make_equations(self):    

        if not self.G.is_planar:
            raise ValueError('Circuit topology is not planar')

        if self.cct.is_superposition:
            raise ValueError('Circuit has a mixture of ac/dc/transient sources')
        
        # This is work in progress to do mesh analysis.

        loops = self.loops()
        Nloops = len(loops)

        mesh_currents = [Iname('I_%d' % (m + 1), self.kind) for m in range(Nloops)]
        equations = {}

        for m, loop in enumerate(loops):

            result = Vtype(self.kind)(0)

            loop1 = loop.copy()
            loop1.append(loop1[0])
            for j in range(len(loop1) - 1):

                elt = self.G.component(loop1[j], loop1[j + 1])
                if elt is None:
                    continue

                if elt.is_current_source:
                    raise ValueError('TODO: handle current source in loop')
                
                # Map node names to equipotential node names.
                nodenames = [self.cct.node_map[nodename] for nodename in elt.nodenames]

                is_reversed = nodenames[0] == loop1[j] and nodenames[1] == loop1[j + 1]
                
                if is_reversed:
                    current = -mesh_currents[m]
                else:
                    current = mesh_currents[m]                    

                for n, loop2 in enumerate(loops):

                    if loop2 == loop:
                        continue

                    loop2 = loop2.copy()
                    loop2.append(loop2[0])

                    if nodenames[0] not in loop2 or nodenames[1] not in loop2:
                        continue

                    # Determine polarity of cpt.
                    for l in range(len(loop2) - 1):
                        if (nodenames[0] == loop2[l] and
                            nodenames[1] == loop2[l + 1]):
                            current -= mesh_currents[n]
                            break
                        elif (nodenames[1] == loop2[l] and
                            nodenames[0] == loop2[l + 1]):
                            current += mesh_currents[n]
                            break

                v = elt.cpt.v_equation(current, self.kind)
                if elt.is_voltage_source and is_reversed:
                    v = -v
                result += v

            equations[mesh_currents[m]] = (result, expr(0))

        return equations

    def mesh_equations_list(self):
        """Return mesh equations as a list."""

        result = ExprList()

        for current, (lhs, rhs) in self._equations.items():
            result.append(equation(lhs, rhs))
        return result        

    def mesh_equations(self):
        """Return mesh equations as a dict keyed by the mesh current."""

        result = ExprDict()

        for current, (lhs, rhs) in self._equations.items():       
            result[current] = equation(lhs, rhs)
        return result

    def _analyse(self):

        if self.kind == 'time':
            raise ValueError('Cannot put time domain equations into matrix form')

        subsdict = {}
        for m, i in enumerate(self._unknowns):
            subsdict[i.expr] = 'X_X%d' % m

        exprs = []
        for node, (lhs, rhs) in self._equations.items():
            lhs = lhs.subs(subsdict).expr.expand()
            rhs = rhs.subs(subsdict).expr.expand()            
            exprs.append(lhs - rhs)
            
        y = []
        for y1 in self._y:
            y.append(y1.subs(subsdict).expr);
        
        A, b = sym.linear_eq_to_matrix(exprs, *y)

        y = [y1.expr for y1 in self._y]
        return SystemEquations(A, b, y)        

    @property
    def A(self):
        """Return A matrix where A y = b."""

        if not hasattr(self, '_sys'):
            self._sys = self._analyse()
        return matrix(self._sys.A)

    @property
    def b(self):
        """Return b vector where A y = b."""        

        if not hasattr(self, '_sys'):
            self._sys = self._analyse()
        return matrix(self._sys.b)        

    @property
    def y(self):
        """Return y vector where A y = b."""
        return self._y
    
    def matrix_equations(self, form='default', invert=False):
        """Return the equations in matrix form.

        Forms can be:
         A y = b
         b = A y
         Ainv b = y
         y = Ainv b

        If `invert` is True, evaluate the matrix inverse."""


        if not hasattr(self, '_sys'):
            self._sys = self._analyse()
        return self._sys.format(form, invert)
    
from .expr import ExprList, ExprDict, expr    
from .current import Iname
from .voltage import Vtype
from .matrix import matrix



from lcapy import Circuit
from numpy import linspace
from matplotlib.pyplot import savefig

cct = Circuit('VRC1.sch')

tv = linspace(0, 1, 100)

results = cct.sim(tv, integrator='backward-euler')

ax = cct.R1.v.plot(tv, label='analytic')
ax.plot(tv, results.R1.v, label='simulated')
ax.legend()

savefig('sim2be.png')



import sympy as sym
from .expr import expr

class SystemEquations(object):

    def __init__(self, A, b, y):
        """Represent a system of equations as A y = b."""

        self.A = A
        self.b = sym.Matrix(b)
        self.y = sym.Matrix(y)

    @property
    def Ainv(self):

        if hasattr(self, '_Ainv'):
            return self._Ainv
        self._Ainv = self.A.inv()
        return self._Ainv        
        
    def format(self, form='A y = b', invert=False):
        """Forms can be:
         A y = b
         b = A y
         Ainv b = y
         y = Ainv b

        If `invert` is True, the A matrix is inverted."""

        if form == 'A y = b':
            return expr(sym.Eq(sym.MatMul(self.A, self.y), self.b), evaluate=False)

        elif form == 'b = A y':
            return expr(sym.Eq(self.b, sym.MatMul(self.A, self.y)), evaluate=False)

        elif form in ('y = Ainv b', 'default'):
            if invert:
                return expr(sym.Eq(self.y, sym.MatMul(self.Ainv, self.b),
                                   evaluate=False))
            
            return expr(sym.Eq(self.y, sym.MatMul(sym.Pow(self.A, -1), self.b),
                               evaluate=False))

        elif form == 'Ainv b = y':
            if invert:
                return expr(sym.Eq(sym.MatMul(self.Ainv, self.b), self.y,
                                   evaluate=False))
            
            return expr(sym.Eq(sym.MatMul(sym.Pow(self.A, -1), self.b), self.y,
                               evaluate=False))
        else:
            raise ValueError('Unknown form %s' % form)
        



"""This module performs network transformations.

Copyright 2020 Michael Hayes, UCECE

"""

from .oneport import R, Y, Z


def Z_wye_to_delta(Z1, Z2, Z3):
    """Perform wye to delta transformation of three impedances.

    This is equivalent to a tee-pi transform or a star-mesh transform."""

    N = Z1 * Z2 + Z1 * Z3 + Z2 * Z3
    
    Za = N / Z1
    Zb = N / Z2
    Zc = N / Z3

    return Za, Zb, Zc


def Z_delta_to_wye(Za, Zb, Zc):
    """Perform wye to delta transformation of three impedances.

    This is equivalent to a pi-tee transform or a mesh-star transform."""    

    D = Za + Zb + Zc
    
    Z1 = Zb * Zc / D
    Z2 = Za * Zc / D
    Z3 = Za * Zb / D

    return Z1, Z2, Z3


def Y_delta_to_wye(Ya, Yb, Yc):
    """Perform wye to delta transformation of three admittances.
        
    This is equivalent to a tee-pi transform or a star-mesh transform."""

    N = Ya * Yb + Ya * Yc + Yb * Yc
    
    Y1 = N / Ya
    Y2 = N / Yb
    Y3 = N / Yc

    return Y1, Y2, Y3


def Y_wye_to_delta(Y1, Y2, Y3):
    """Perform wye to delta transformation of three admittances.

    This is equivalent to a pi-tee transform or a mesh-star transform."""    

    D = Y1 + Y2 + Y3
    
    Ya = Y2 * Y3 / D
    Yb = Y1 * Y3 / D
    Yc = Y1 * Y2 / D

    return Ya, Yb, Yc


def wye_to_delta(cpt1, cpt2, cpt3):

    if cpt1.is_impedance:    
        return Z_wye_to_delta(cpt1, cpt2, cpt3)

    if cpt1.is_admittance:        
        return Y_wye_to_delta(cpt1, cpt2, cpt3)    

    if isinstance(cpt1, Z):
        return [Z(cpt1) for cpt1 in Z_wye_to_delta(cpt1.Z, cpt2.Z, cpt3.Z)]

    if isinstance(cpt1, R):
        return [R(R1) for R1 in Z_wye_to_delta(cpt1.R, cpt2.R, cpt3.R)]

    if isinstance(cpt1, Y):
        return [Y(Y1) for Y1 in Y_wye_to_delta(cpt1.Y, cpt2.Y, cpt3.Y)]        

    raise ValueError('Unhandled class %s' % cpt1.__class__.name)

    
def delta_to_wye(cpt1, cpt2, cpt3):

    if cpt1.is_impedance:
        return Z_delta_to_wye(cpt1, cpt2, cpt3)

    if cpt1.is_admittance:    
        return Y_delta_to_wye(cpt1, cpt2, cpt3)    

    if isinstance(cpt1, Z):
        return [Z(cpt1) for cpt1 in Z_delta_to_wye(cpt1.Z, cpt2.Z, cpt3.Z)]

    if isinstance(cpt1, R):
        return [R(R1) for R1 in Z_delta_to_wye(cpt1.R, cpt2.R, cpt3.R)]

    if isinstance(cpt1, Y):
        return [Y(Y1) for Y1 in Y_delta_to_wye(cpt1.Y, cpt2.Y, cpt3.Y)]        

    raise ValueError('Unhandled class %s' % cpt1.__class__.name)

    




from lcapy import *
from numpy import logspace
from matplotlib.pyplot import figure, savefig

N = R(10) | C(1e-4) | L(1e-3)

vf = logspace(0, 5, 400)
Z = N.Z(f).evaluate(vf)

fig = figure()
ax = fig.add_subplot(111)
ax.loglog(vf, abs(Z), linewidth=2)
ax.set_xlabel('Frequency (Hz)')
ax.set_ylabel('Impedance (ohms)')
ax.grid(True)


savefig('parallel-RLC3-Z.png')



"""This module provides the Units class for simplification of units.
It should be rolled into SymPy.  It can perform simplification of
units, e.g., volts / amperes -> ohms.

Copyright 2020--2021 Michael Hayes, UCECE

"""

import sympy.physics.units as u
from sympy.physics.units.systems.si import dimsys_SI
from sympy.physics.units.systems import SI 
from sympy.physics.units import UnitSystem


class Units(object):

    def __init__(self, unit_system="SI"):

        self.unit_system = UnitSystem.get_unit_system(unit_system)
        self.dim_sys = self.unit_system.get_dimension_system()
        self._mapping = {}
        
        for i in u.__dict__:          
            unit = getattr(u, i)
            if not isinstance(unit, u.Quantity):
                continue

            key = self._makekey(unit)            

            # Use earlier defined units
            
            if key not in self._mapping:
                self._mapping[key] = unit

        # Remove entry for no units.
        self._mapping.pop(self._makekey(1))

    def _get_dependencies(self, unit):
            
        dim = self.unit_system.get_dimensional_expr(unit)
        return self.dim_sys.get_dimensional_dependencies(dim)
            
    def _makekey(self, unit):
            
        deps = self._get_dependencies(unit)
        key = tuple([deps.get(str(dim.name)) for dim in self.dim_sys.base_dims])
        return key
            
    def simplify_units(self, unit):

        key = self._makekey(unit)
        if not key in self._mapping:
            return unit
        return self._mapping[key]

    def simplify(self, expr):

        value, unit = self.as_value_unit(expr)
        return value * self.simplify_units(unit)

    def as_value_unit(self, expr):
        return as_value_unit(expr)

    
def as_value_unit(expr):

    if isinstance(expr, u.Quantity):
        return 1, expr

    if not expr.has(u.Quantity):
        return expr, 1
    
    defs = {x: 1 for x in expr.args if not x.has(u.Quantity)}
    unit = expr.subs(defs)

    return expr / unit, unit


units = Units()



"""This module defines special symbols f, s, t, and omega.

Copyright 2014--2021 Michael Hayes, UCECE

"""

# The following are Lcapy expressions.  For example, f is an
# FourierDomainExpression object wrapping the fsym symbol.
from .fexpr import f
from .texpr import t
from .sexpr import s
from .omegaexpr import omega, omega0

# Import common SymPy symbols.
from .sym import pi, j, oo, inf, one

jomega = j * omega
jw = jomega

jomega0 = j * omega0
jw0 = jomega0

# This represents an arbitrary angular frequency
w = omega
# This represents a specific angular frequency and is assumed to be positive
w0 = omega0
